import logging
import os
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np

from .acquisition_fun import EI, MGFI, PI, UCB
from .bayes_opt import BO, AnnealingBO, NoisyBO, ParallelBO
from .mobo import MOBO
from .search_space import (
    Bool,
    BoolSpace,
    Discrete,
    DiscreteSpace,
    Integer,
    IntegerSpace,
    Ordinal,
    Real,
    RealSpace,
    SearchSpace,
)
from .solution import Solution
from .surrogate import GaussianProcess, RandomForest, trend

__all__ = [
    "BO",
    "ParallelBO",
    "NoisyBO",
    "AnnealingBO",
    "MOBO",
    "Solution",
    "RandomForest",
    "GaussianProcess",
    "trend",
    "SearchSpace",
    "IntegerSpace",
    "RealSpace",
    "BoolSpace",
    "DiscreteSpace",
    "EI",
    "UCB",
    "PI",
    "MGFI",
    "RandomForest",
    "fmin",
    "Integer",
    "Ordinal",
    "Real",
    "Bool",
    "Discrete",
]

# To use `dill` for the pickling, which works for
# much more python objects
os.environ["LOKY_PICKLER"] = "dill"

verbose = {False: logging.NOTSET, "DEBUG": logging.DEBUG, "INFO": logging.INFO}

Vector = List[float]
Matrix = List[Vector]

# TODO: implement the code for `callback`, `xtol`, `ftol`, and `kwargs`
def fmin(
    func: Callable,
    lower: Union[float, Vector],
    upper: Union[float, Vector],
    x0: Union[int, Matrix, np.ndarray, None] = None,
    y0: Union[Vector, None] = None,
    n_point: int = 1,
    args: Tuple = (),
    xtol: float = 1e-4,
    ftol: float = 1e-4,
    max_FEs: Optional[int] = None,
    verbose: Optional[bool] = False,
    callback: Optional[Callable] = None,
    seed: Optional[int] = None,
    **kwargs,
) -> Tuple[Vector, float, int, int, List[np.ndarray]]:
    """Minimize a function using the Bayesian Optimization algorithm, which only uses
    function values, not derivatives or second derivatives. This function maintains an
    interface similar to `scipy.optimize.fmin`. Hereafter, we use the following customized
    types to describe the usage:

    - Vector = List[float]
    - Matrix = List[Vector]

    Parameters
    ----------
    func : Callable
        The objective function to be minimized.
    lower : Union[float, Vector]
        The lower bound of search variables, from which the search dimension is inferred.
        When it is not a `float`, it must have the same length as `upper`.
    upper : Union[float, Vector]
        The upper bound of search variables, from which the search dimension is inferred.
        When it is not a `float`, it must have the same length as `lower`.
    x0 : Union[int, Matrix, None], optional
        When it takes integer values, it specifies the number of initial sample points to
        take; When it is a 2d numpy array (or a nested list that contains the same data),
        it provides the initial sample points.
    y0 : Union[Vector, None], optional
        When `x0` is a 2d numpy array, it contains the function values pertaining to `x0`;
        Otherwise, it is ignored. It is by default None.
    n_point : int, optional
        The number of trial points generated in each iteration, by default 1
    args : Tuple, optional
        Extra arguments passed to `func`, i.e., ``func(x, *args)``.
    xtol : float, optional
        Absolute error in xopt between iterations that is acceptable for convergence,
        by default 1e-4.
    ftol : float, optional
        Absolute error in func(xopt) between iterations that is acceptable for convergence,
        by default 1e-4.
    max_FEs : Optional[int], optional
        Maximal number of function evaluations to make, by default None.
    verbose : Optional[bool], optional
        Verbosity of the output, by default False.
    callback : Optional[Callable], optional
        Called after each iteration, as `callback(X)`, where `X` is the current parameter
        vectors, by default None.
    seed : Optional[int], optional
        Seeding the random number generator in `numpy`, by default None.

    Returns
    -------
    Tuple[Vector, float, int, int, List[np.ndarray]]
        The return value contains:
        - The best-so-far parameter vector
        - The best-so-far function value
        - The number of iterations
        - The number of function evaluations
        - A list of trial points/vectors generated in every iteration

    Examples
    --------
    >>> def f(x):
    ...     return sum(x ** 2)
    >>> from bayes_optim import fmin
    >>>
    >>> minimum = fmin(f, [-5] * 2, [5] * 2, max_FEs=30, seed=42)
    Optimization terminated successfully.
            Current function value: 0.007165794451494286
            Iterations: 21
            Function evaluations: 30
    >>> minimum[0]
    0.007165794451494286
    """
    if seed is not None:
        _state = np.random.get_state()
        np.random.seed(seed)

    obj_func = lambda x: func(x, *args)

    if isinstance(lower, float) and isinstance(upper, float):
        dim = 1
        space = RealSpace([lower, upper])
    elif isinstance(lower, list) and isinstance(upper, list):
        assert len(lower) == len(upper)
        dim = len(lower)
        space = RealSpace(list(zip(lower, upper)))
        lower, upper = np.array(lower), np.array(upper)

    mean = trend.constant_trend(dim, beta=0)
    thetaL = 1e-10 * (upper - lower) * np.ones(dim)
    thetaU = 10 * (upper - lower) * np.ones(dim)
    theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL
    model = GaussianProcess(
        mean=mean,
        corr="squared_exponential",
        theta0=theta0,
        thetaL=thetaL,
        thetaU=thetaU,
        nugget=0,
        noise_estim=False,
        optimizer="BFGS",
        wait_iter=3,
        random_start=dim,
        likelihood="concentrated",
        eval_budget=100 * dim,
    )

    # set up the warm-starting and DoE size
    if isinstance(x0, int):
        DoE_size = x0
        warm_data = ()
    elif hasattr(x0, "__iter__"):
        DoE_size = None
        if y0 is None:
            y0 = [obj_func(_) for _ in x0]
        warm_data = (x0, y0)
    else:
        DoE_size = None
        warm_data = ()

    _BO = BO if n_point == 1 else ParallelBO
    opt = _BO(
        search_space=space,
        obj_fun=obj_func,
        model=model,
        DoE_size=DoE_size,
        warm_data=warm_data,
        eval_type="list",
        max_FEs=max_FEs,
        verbose=verbose,
        n_point=n_point,
    )
    opt.run()

    if seed is not None:
        np.random.set_state(_state)

    N, n = opt.DoE_size, opt.n_point
    _data, data = opt.data[:N, :], opt.data[N:, :]

    data_per_iteration = [np.asarray(data[:N, :])]
    data_per_iteration += [
        np.asarray(data[(i * n) : ((i + 1) * n), :]) for i in range(opt.iter_count - 1)
    ]

    print(
        "Optimization terminated successfully.\n"
        "        Current function value: {}\n"
        "        Iterations: {}\n"
        "        Function evaluations: {}\n".format(
            opt.xopt.fitness, opt.iter_count, opt.eval_count
        )
    )
    return opt.xopt, opt.xopt.fitness, opt.iter_count, opt.eval_count, data_per_iteration



import warnings
from abc import ABC, abstractmethod

import numpy as np
from numpy import exp, sqrt
from scipy.stats import norm

__authors__ = ["Hao Wang"]

# TODO: add annotations
# TODO: implement noisy handling infill criteria, e.g., EQI (expected quantile improvement)
# TODO: perphaps also enable acquisition function engineering here?
# TODO: multi-objective extension


class AcquisitionFunction(ABC):
    def __init__(self, model=None, minimize=True):
        self.model = model
        self.minimize = minimize

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        if model is not None:
            self._model = model
            assert hasattr(self._model, "predict")
        else:
            self._model = None

    @abstractmethod
    def __call__(self, X):
        raise NotImplementedError

    def _predict(self, X):
        y_hat, sd2 = self._model.predict(X, eval_MSE=True)
        sd = sqrt(sd2)
        if not self.minimize:
            y_hat = -y_hat
        return y_hat, sd

    def _gradient(self, X):
        X_ = np.array(X, dtype=float)
        y_dx, sd2_dx = self._model.gradient(X_)
        if not self.minimize:
            y_dx = -y_dx
        return y_dx, sd2_dx

    def check_X(self, X):
        """Keep input as '2D' object"""
        return np.atleast_2d(X)


class ImprovementBased(AcquisitionFunction):
    def __init__(self, plugin=None, **kwargs):
        super().__init__(**kwargs)
        self.plugin = plugin

    @property
    def plugin(self):
        return self._plugin

    @plugin.setter
    def plugin(self, plugin):
        if plugin is None:
            if self._model is not None:
                self._plugin = (
                    np.min(self._model.y) if self.minimize else -1.0 * np.max(self._model.y)
                )
            else:
                self._plugin = None
        else:
            self._plugin = plugin if self.minimize else -1.0 * plugin


class UCB(AcquisitionFunction):
    def __init__(self, alpha=0.5, **kwargs):
        """Upper Confidence Bound"""
        super().__init__(**kwargs)
        self.alpha = alpha

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        assert alpha > 0
        self._alpha = alpha

    def __call__(self, X, return_dx=False):
        X = self.check_X(X)
        n_sample = X.shape[0]
        y_hat, sd = self._predict(X)

        try:
            f_value = y_hat + self.alpha * sd
            if n_sample == 1:
                f_value = sum(f_value)
        except Exception:  # in case of numerical errors
            f_value = 0

        if return_dx:
            y_dx, sd2_dx = self._gradient(X)
            sd_dx = sd2_dx / (2.0 * sd)
            try:
                f_dx = y_dx + self.alpha * sd_dx
            except Exception:
                f_dx = np.zeros((len(X[0]), 1))
            return f_value, f_dx
        return f_value


class EI(ImprovementBased):
    def __call__(self, X, return_dx=False):
        X = self.check_X(X)
        n_sample = X.shape[0]
        y_hat, sd = self._predict(X)

        # if the Kriging variance is to small
        # TODO: check the rationale of 1e-6 and why the ratio if intended
        if hasattr(self._model, "sigma2"):
            if sd / np.sqrt(self._model.sigma2) < 1e-6:
                return (0, np.zeros((len(X[0]), 1))) if return_dx else 0
        else:
            # TODO: implement a counterpart of 'sigma2' for randomforest
            # or simply put a try...except around the code below
            if sd < 1e-10:
                return (0, np.zeros((len(X[0]), 1))) if return_dx else 0
        try:
            xcr_ = self._plugin - y_hat
            xcr = xcr_ / sd
            xcr_prob, xcr_dens = norm.cdf(xcr), norm.pdf(xcr)
            f_value = xcr_ * xcr_prob + sd * xcr_dens
            if n_sample == 1:
                f_value = sum(f_value)
        except Exception:  # in case of numerical errors
            # IMPORTANT: always keep the output in the same type
            f_value = 0

        if return_dx:
            y_dx, sd2_dx = self._gradient(X)
            sd_dx = sd2_dx / (2.0 * sd)
            try:
                f_dx = -y_dx * xcr_prob + sd_dx * xcr_dens
            except Exception:
                f_dx = np.zeros((len(X[0]), 1))
            return f_value, f_dx
        return f_value


class EpsilonPI(ImprovementBased):
    def __init__(self, epsilon=1e-10, **kwargs):
        """epsilon-Probability of Improvement
        # TODO: verify the implementation
        """
        super(EpsilonPI, self).__init__(**kwargs)
        self.epsilon = epsilon

    @property
    def epsilon(self):
        return self._epsilon

    @epsilon.setter
    def epsilon(self, eps):
        assert eps > 0
        self._epsilon = eps

    def __call__(self, X, return_dx=False):
        X = self.check_X(X)
        y_hat, sd = self._predict(X)

        coef = 1 - self._epsilon if y_hat > 0 else (1 + self._epsilon)
        try:
            xcr_ = self._plugin - coef * y_hat
            xcr = xcr_ / sd
            f_value = norm.cdf(xcr)
        except Exception:
            f_value = 0.0

        if return_dx:
            y_dx, sd2_dx = self._gradient(X)
            sd_dx = sd2_dx / (2.0 * sd)
            try:
                f_dx = -(coef * y_dx + xcr * sd_dx) * norm.pdf(xcr) / sd
            except Exception:
                f_dx = np.zeros((len(X[0]), 1))
            return f_value, f_dx
        return f_value


class PI(EpsilonPI):
    def __init__(self, **kwargs):
        """Probability of Improvement"""
        kwargs.update({"epsilon": 0})
        super().__init__(**kwargs)


class MGFI(ImprovementBased):
    def __init__(self, t=1, **kwargs):
        """Moment-Generating Function of Improvement proposed in SMC'17 paper"""
        super().__init__(**kwargs)
        self.t = t

    @property
    def t(self):
        return self._t

    @t.setter
    def t(self, t):
        assert t > 0
        t = min(t, 22.36)  # bigger `t` value would cause numerical overflow
        self._t = t

    def __call__(self, X, return_dx=False):
        X = self.check_X(X)
        y_hat, sd = self._predict(X)
        n_sample = X.shape[0]

        # if the Kriging variance is to small
        # TODO: check the rationale of 1e-6 and why the ratio if intended
        if np.isclose(sd, 0):
            return (np.array([0.0]), np.zeros((len(X[0]), 1))) if return_dx else 0.0

        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            try:
                y_hat_p = y_hat - self._t * sd ** 2.0
                beta_p = (self._plugin - y_hat_p) / sd
                term = self._t * (self._plugin - y_hat - 1)
                f_ = norm.cdf(beta_p) * exp(term + self._t ** 2.0 * sd ** 2.0 / 2.0)
                if n_sample == 1:
                    f_ = sum(f_)
            except Exception:  # in case of numerical errors
                f_ = 0.0

        if np.isinf(f_):
            f_ = 0.0

        if return_dx:
            y_dx, sd2_dx = self._gradient(X)
            sd_dx = sd2_dx / (2.0 * sd)

            with warnings.catch_warnings():
                warnings.filterwarnings("error")
                try:
                    term = exp(self._t * (self._plugin + self._t * sd ** 2.0 / 2 - y_hat - 1))
                    m_prime_dx = y_dx - 2.0 * self._t * sd * sd_dx
                    beta_p_dx = -(m_prime_dx + beta_p * sd_dx) / sd

                    f_dx = term * (
                        norm.pdf(beta_p) * beta_p_dx
                        + norm.cdf(beta_p) * ((self._t ** 2) * sd * sd_dx - self._t * y_dx)
                    )
                except Exception:
                    f_dx = np.zeros((len(X[0]), 1))
            return f_, f_dx
        return f_


class GEI(ImprovementBased):
    def __init__(self, g=1, **kwargs):
        """Generalized Expected Improvement"""
        super().__init__(**kwargs)
        self.g = g

    @property
    def g(self):
        return self._g

    @g.setter
    def g(self, g):
        g = int(g)
        assert g >= 0
        self._g = g

    def __call__(self, X, return_dx=False):
        # TODO: implement this!!!
        raise NotImplementedError



import sys

import numpy as np

sys.path.insert(0, "../")

from bayes_optim import DiscreteSpace, IntegerSpace, RealSpace
from bayes_optim.search_space.samplers import SCMC

search_space = (
    RealSpace([-5, 5]) * 2 + DiscreteSpace(["A", "B", "C", "D"]) + IntegerSpace([1, 10]) * 2
)


def h(x):
    return np.array([bool(x[2] not in ["A", "B"]), x[4] ** 2 + x[4] - 2])


def g(x):
    return np.array(
        [
            np.sum(x[:2] ** 2) - 1,
            0.25 - np.sum(x[:2] ** 2),
            x[3] - 5.1,
        ]
    )


def test_SCMC():
    sampler = SCMC(
        search_space,
        lambda x: np.r_[np.abs(h(x)) if h else [], np.array(g(x)) if g else []],
        tol=1e-1,
    )
    X = sampler.sample(10)
    assert all([np.all(np.isclose(h(x), 0, atol=1e-1)) for x in X])
    assert all([np.all(g(x) <= 0) for x in X])


def test_search_space_sampling():
    X = search_space.sample(10, h=h, g=g)
    assert all([np.all(np.isclose(h(x), 0, atol=1e-2)) for x in X])
    assert all([np.all(g(x) <= 0) for x in X])



from __future__ import annotations

import re
import sys
import warnings
from abc import ABC
from copy import copy, deepcopy
from itertools import chain, combinations
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
from numpy.random import randint
from scipy.special import logit

__authors__ = ["Hao Wang"]

MAX = sys.float_info.max
TRANS = {
    "linear": [lambda x: x, [-MAX, MAX]],
    "log": [np.log, [1e-300, MAX]],
    "log10": [np.log10, [1e-300, MAX]],
    "logit": [logit, [1e-300, 1]],
    "bilog": [lambda x: np.sign(x) * np.log(1 + np.abs(x)), [-MAX, MAX]],
}
INV_TRANS = {
    "linear": lambda x: x,
    "log": np.exp,
    "log10": lambda x: np.power(10, x),
    "logit": lambda x: 1 / (1 + np.exp(-x)),
    "bilog": lambda x: np.sign(x) * (np.exp(np.abs(x)) - 1),
}


class Variable(ABC):
    """Base class for decision variables"""

    def __init__(
        self,
        bounds: List[int, float, str],
        name: str,
        default_value: Union[int, float, str] = None,
        conditions: str = None,
        action: Union[callable, int, float, str] = lambda x: x,
    ):
        """Base class for decision variables

        Parameters
        ----------
        bounds : List[int, float, str]
            a list/tuple giving the range of the variable.
                * For `Real`, `Integer`: (lower, upper)
                * For `Ordinal` and `Discrete`: (value1, value2, ...)
        name : str
            variable name
        default_value : Union[int, float, str], optional
            default value, by default None
        conditions : str, optional
            a string specifying the condition on which the variable is problematic, e.g.,
            being either invalid or ineffective, by default None. The variable name in
            this string should be quoted as `var name`. Also, you could use multiple
            variables and logic conjunctions/disjunctions therein.
            Example: "`var1` == True and `var2` == 2"
        action : Union[callable, int, float, str], optional
            the action to take when `condition` evaluates to True, by default `lambda x: x`.
            It can be simply a fixed value to which the variable will be set, or a callable
            that determines which value to take.
        """
        if len(bounds) > 0 and isinstance(bounds[0], list):
            bounds = bounds[0]
        self.name: str = name
        self.bounds = tuple(bounds)
        self.set_default_value(default_value)
        self.add_conditions(conditions, action)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        msg = f"{self.name} -> {type(self).__name__} | range: {self.bounds}"
        if self.default_value is not None:
            msg += f" | default: {self.default_value}"
        return msg

    def __contains__(self, x: Union[float, str, object]) -> bool:
        pass

    def __eq__(self, var: Variable) -> bool:
        return (
            self.__class__ == type(var)
            and self.bounds == var.bounds
            and self.default_value == var.default_value
            and self.name == var.name
            and self._conditions == var._conditions
        )  # TODO: verify this!

    def __ne__(self, var: Variable) -> bool:
        return not self.__eq__(var)

    def copyfrom(self, var: Variable):
        self.__dict__.update(**deepcopy(var.__dict__))

    def set_default_value(self, value):
        """validate the default value first"""
        if value is not None:
            assert self.__contains__(value)
        self.default_value = value

    def add_conditions(self, conditions: str, action: Union[callable, int, float, str]):
        self._conditions = None
        if conditions is not None:
            self.var_in_conditions = re.findall(r"`([^`]*)`", conditions)
            for i, var_name in enumerate(self.var_in_conditions):
                conditions = conditions.replace(f"`{var_name}`", f"#{i}")
            self._conditions = conditions

        if isinstance(action, (int, float, str)):
            self._action = lambda x: action
        elif hasattr(action, "__call__"):
            self._action = action

    @property
    def conditions(self):
        if self._conditions is None:
            return None
        out = copy(self._conditions)
        for i, var_name in enumerate(self.var_in_conditions):
            out = out.replace(f"#{i}", f"{var_name}")
        return out

    @property
    def action(self):
        return self._action


class Real(Variable):
    """Real-valued variable taking its value in a continuum"""

    def __init__(
        self,
        bounds: Tuple[float, float],
        name: str = "r",
        default_value: float = None,
        precision: int = None,
        scale: str = "linear",
        **kwargs,
    ):
        """Real-valued variable taking its value in a continuum

        Parameters
        ----------
        bounds : [Tuple[float, float]
            the lower and upper bound
        name : str, optional
            the variable name, by default 'r'
        default_value : float, optional
            the default value, by default None
        precision : int, optional
            the number of digits after decimal, by default None
        scale : str, optional
            the scale on which uniform sampling is performed, by default 'linear'
        """
        assert bounds[0] < bounds[1]
        assert scale in TRANS.keys()
        assert precision is None or isinstance(precision, int)
        super().__init__(bounds, name, default_value, **kwargs)
        self.precision: int = precision
        self.scale = scale

    def __hash__(self):
        return hash((self.name, self.bounds, self.default_value, self.precision, self.scale))

    def __str__(self):
        msg = super().__str__()
        if self.precision:
            msg += f" | precision: .{self.precision}f"
        msg += f" | scale: {self.scale}"
        return msg

    def __contains__(self, x: Union[float, str]) -> bool:
        return self.bounds[0] <= x <= self.bounds[1]

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, scale):
        if scale is None:
            scale = "linear"

        assert scale in TRANS.keys()
        self._scale: str = scale
        self._trans: Callable = TRANS[scale][0]
        self._inv_trans: Callable = INV_TRANS[scale]
        _range = TRANS[scale][1]
        bounds = list(self.bounds)

        if (bounds[0] < _range[0]) or (bounds[0] > _range[1]):
            bounds[0] = _range[0]
            warnings.warn(
                f"lower bound {bounds[0]} not in the working "
                f"range of the given scale {self._scale} is set to the default value {_range[0]}"
            )
            # raise ValueError(
            #     f"lower bound {self.bounds[0]} not in the working "
            #     f"range of the given scale {self._scale}"
            # )

        if (bounds[1] < _range[0]) or (bounds[1] > _range[1]):
            bounds[1] = _range[1]
            warnings.warn(
                f"upper bound {bounds[1]} not in the working "
                f"range of the given scale {self._scale} is set to the default value {_range[1]}"
            )
            # raise ValueError(
            #     f"upper bound {self.bounds[1]} is invalid for the given scale {self._scale}"
            # )
        self.bounds = tuple(bounds)
        self._bounds_transformed = self._trans(self.bounds)

    def to_linear_scale(self, X):
        return X if self.scale == "linear" else self._inv_trans(X)

    def round(self, X):
        """Round the real-valued components of `X` to the
        corresponding numerical precision, if given
        """
        X = deepcopy(X)
        if self.precision is not None:
            X = np.round(X, self.precision)
        return X


class _Discrete(Variable):
    """Represents Integer, Ordinal, Bool, and Discrete"""

    def __init__(self, bounds, *args, **kwargs):
        # get rid of duplicated levels
        bounds = list(dict.fromkeys(bounds))
        # map discrete values (bounds) to integers for sampling
        self._map_func: callable = None
        self._size: int = None
        super().__init__(bounds, *args, **kwargs)

    def __contains__(self, x: Union[int, str]) -> bool:
        return x in self.bounds

    def __hash__(self):
        return hash((self.name, self.bounds, self.default_value))

    def sample(self, N: int = 1, **kwargs) -> List:
        return list(map(self._map_func, randint(0, self._size, N)))


class Discrete(_Discrete):
    """Discrete variable, whose values should come with a linear order"""

    def __init__(self, bounds, name: str = "d", default_value: Union[int, str] = None, **kwargs):
        super().__init__(bounds, name, default_value, **kwargs)
        self._map_func = lambda i: self.bounds[i]
        self._size = len(self.bounds)


class Subset(Discrete):
    """A discrete variable created by enumerating all subsets of the input `bounds`"""

    def __init__(self, bounds, name: str = "s", default_value: Union[int, str] = None, **kwargs):
        self._bounds = bounds
        bounds = list(
            chain.from_iterable(map(lambda r: combinations(bounds, r), range(1, len(bounds) + 1)))
        )
        super().__init__(bounds, name, default_value, **kwargs)

    def __str__(self):
        msg = f"{self.name} -> {type(self).__name__} | range: 2 ^ {self._bounds}"
        if self.default_value is not None:
            msg += f" | default: {self.default_value}"
        return msg


class Ordinal(_Discrete):
    """A generic ordinal variable, whose values should come with a linear order"""

    def __init__(self, bounds, name: str = "ordinal", default_value: int = None, **kwargs):
        super().__init__(bounds, name, default_value, **kwargs)
        self._map_func = lambda i: self.bounds[i]
        self._size = len(self.bounds)


class Integer(_Discrete):
    """Integer variable"""

    def __init__(
        self,
        bounds: Tuple[int],
        name: str = "i",
        default_value: int = None,
        step: Optional[Union[int, float]] = 1,
    ):
        super().__init__(bounds, name, default_value)
        assert len(self.bounds) == 2
        assert self.bounds[0] < self.bounds[1]
        assert all(map(lambda x: isinstance(x, (int, float)), self.bounds))
        self.step = step
        self._map_func = lambda i: self.bounds[0] + i * self.step
        self._size = int(np.floor((self.bounds[1] - self.bounds[0]) / self.step) + 1)

    def __contains__(self, x: int) -> bool:
        return self.bounds[0] <= x <= self.bounds[1]

    def __hash__(self):
        return hash((self.name, self.bounds, self.default_value, self.step))

    def __str__(self):
        msg = super().__str__()
        msg += f" | step: {self.step}"
        return msg


class Bool(_Discrete):
    """Boolean variable"""

    def __init__(self, name: str = "bool", default_value: int = True, **kwargs):
        # NOTE: remove `bounds` if it presents in the input
        kwargs.pop("bounds", None)
        assert default_value is None or isinstance(default_value, bool)
        super().__init__((False, True), name, default_value, **kwargs)
        self._map_func = bool
        self._size = 2



from __future__ import annotations

from collections import OrderedDict
from typing import List, Union

import numpy as np
from joblib import Parallel, delayed
from numpy import array
from sklearn.ensemble import RandomForestRegressor
from sklearn.ensemble._base import _partition_estimators
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils.validation import check_is_fitted

from ..solution import Solution

__authors__ = ["Hao Wang"]


class SurrogateAggregation(object):
    """Linear aggregation of surrogate models used for multi-obvjective optimization"""

    def __init__(self, surrogates, aggregation="WS", **kwargs):
        self.surrogates = surrogates
        self.N = len(self.surrogates)
        self.aggregation = aggregation
        self.weights = np.asarray(kwargs["weights"], dtype="float").ravel()

        assert self.aggregation in ["WS", "Tchebycheff"]

    def fit(self, X, y):
        pass

    def predict(self, X, eval_MSE=False):
        if eval_MSE:
            y_hat_, MSE_ = list(zip(*[s.predict(X, eval_MSE=True) for s in self.surrogates]))
            y_hat_ = np.atleast_2d([_.ravel() for _ in y_hat_])
            MSE_ = np.atleast_2d([_.ravel() for _ in MSE_])
        else:
            y_hat_ = np.atleast_2d([_.predict(X, eval_MSE=False).ravel() for _ in self.surrogates])

        if self.aggregation == "WS":
            y_hat = self.weights.dot(y_hat_)
            if eval_MSE:
                MSE = (self.weights ** 2.0).dot(MSE_)

        elif self.aggregation == "Tchebycheff":
            # TODO: implement this part
            pass

        return (y_hat, MSE) if eval_MSE else y_hat

    def gradient(self, X):
        # TODO: this model is not differentiable?
        pass


def _save_prediction(predict, X, index, out):
    """It can't go locally in ForestClassifier or ForestRegressor, because joblib
    complains that it cannot pickle it when placed there.
    """
    out[..., index] = predict(X, check_input=False)


class RandomForest(RandomForestRegressor):
    r"""Extension on the sklearn's `RandomForestRegressor`
    Added functionality:
        1) MSE estimate,
        2) OneHotEncoding to handle categorical variables
    """

    def __init__(
        self,
        n_estimators: int = 100,
        max_features: float = 5 / 6,
        min_samples_leaf: int = 2,
        levels: dict = None,
        **kwargs,
    ):
        """
        parameter
        ---------
        levels : dict, for categorical inputs
            keys: indices of categorical variables
            values: list of levels of categorical variables
        """
        super().__init__(
            n_estimators=n_estimators,
            max_features=max_features,
            min_samples_leaf=min_samples_leaf,
            **kwargs,
        )
        assert isinstance(levels, dict)
        self.levels = levels
        self.is_fitted = False

        # for categorical levels/variable number
        # in the future, maybe implement binary/multi-value split
        if self.levels:
            self._levels = OrderedDict(sorted(levels.items()))
            self._cat_idx = list(self._levels.keys())
            self._categories = list(self._levels.values())
            # encode categorical variables to binary values
            self._enc = OneHotEncoder(categories=self._categories, sparse=False)

    def _check_X(self, X: Union[Solution, List, np.ndarray]) -> Solution:
        X_ = array(X, dtype=object)
        if hasattr(self, "_levels"):
            X_cat = X_[:, self._cat_idx]
            try:
                X_cat = self._enc.transform(X_cat)
            except Exception:
                X_cat = self._enc.fit_transform(X_cat)
            X = np.c_[np.delete(X_, self._cat_idx, 1).astype(float), X_cat]
        return X

    def fit(self, X: Union[Solution, List, np.ndarray], y: np.ndarray):
        self.X = self._check_X(X)
        self.y = y
        self.is_fitted = True
        return super().fit(self.X, self.y)

    def predict(self, X: Union[Solution, List, np.ndarray], eval_MSE=False) -> np.ndarray:
        """Predict regression target for `X`.
        The predicted regression target of an input sample is computed as the
        mean predicted regression targets of the trees in the forest.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.
        Returns
        -------
        y : ndarray of shape (n_samples,) or (n_samples, n_outputs)
            The predicted values.
        """
        check_is_fitted(self)
        # check data
        X = self._check_X(X)
        X = self._validate_X_predict(X)
        # assign chunk of trees to jobs
        n_jobs, _, _ = _partition_estimators(self.n_estimators, self.n_jobs)
        # storing the output of every estimator since those are required to estimate the MSE
        y_hat_all = (
            np.zeros((X.shape[0], self.n_outputs_, self.n_estimators), dtype=np.float64)
            if self.n_outputs_ > 1
            else np.zeros((X.shape[0], self.n_estimators), dtype=np.float64)
        )
        # parallel loop
        Parallel(n_jobs=n_jobs, verbose=self.verbose, backend="threading")(
            delayed(_save_prediction)(e.predict, X, i, y_hat_all)
            for i, e in enumerate(self.estimators_)
        )
        y_hat = np.mean(y_hat_all, axis=-1)

        if eval_MSE:
            # TODO: implement the jackknife estimate of variance
            MSE_hat = np.std(y_hat_all, axis=-1, ddof=1) ** 2.0
        return (y_hat, MSE_hat) if eval_MSE else y_hat



import logging
import sys
from copy import copy
from typing import Callable, Dict, List, Union

import numpy as np
from scipy.linalg import solve_triangular

from ..misc import LoggerFormatter, handle_box_constraint
from ..search_space import RealSpace, SearchSpace
from ..utils import dynamic_penalty, set_bounds

Vector = List[float]
Matrix = List[Vector]

__authors__ = ["Hao Wang"]


class OnePlusOne_CMA(object):
    """(1+1)-CMA-ES"""

    def __init__(
        self,
        search_space: SearchSpace,
        obj_fun: Callable,
        args: Dict = None,
        h: Callable = None,
        g: Callable = None,
        x0: Union[str, Vector, np.ndarray] = None,
        sigma0: Union[float] = None,
        C0: Union[Matrix, np.ndarray] = None,
        ftarget: Union[int, float] = None,
        max_FEs: Union[int, str] = np.inf,
        minimize: bool = True,
        n_restart: int = 0,
        xtol: float = 1e-4,
        ftol: float = 1e-4,
        verbose: bool = False,
        logger: str = None,
        random_seed: int = 42,
        **kwargs,
    ):
        """Hereafter, we use the following customized
        types to describe the usage:

        - Vector = List[float]
        - Matrix = List[Vector]

        Parameters
        ----------
        dim : int
            Dimensionality of the search space.
        obj_fun : Callable
            The objective function to be minimized.
        args: Tuple
            The extra parameters passed to function `obj_fun`.
        h : Callable, optional
            The equality constraint function, by default None.
        g : Callable, optional
            The inequality constraint function, by default None.
        x0 : Union[str, Vector, np.ndarray], optional
            The initial guess (by default None) which must fall between lower
            and upper bounds, if non-infinite values are provided for `lb` and
            `ub`. Note that, `x0` must be provided when `lb` and `ub` both
            take infinite values.
        sigma0 : Union[float], optional
            The initial step size, by default None
        C0 : Union[Matrix, np.ndarray], optional
            The initial covariance matrix which must be positive definite,
            by default None. Any non-positive definite input will be ignored.
        lb : Union[float, str, Vector, np.ndarray], optional
            The lower bound of search variables. When it is not a `float`,
            it must have the same length as `upper`, by default `-np.inf`.
        ub : Union[float, str, Vector, np.ndarray], optional
            The upper bound of search variables. When it is not a `float`,
            it must have the same length as `lower`, by default `np.inf`.
        ftarget : Union[int, float], optional
            The target value to hit, by default None.
        max_FEs : Union[int, str], optional
            Maximal number of function evaluations to make, by default `np.inf`.
        minimize : bool, optional
            To minimize or maximize, by default True.
        xtol : float, optional
            Absolute error in xopt between iterations that is acceptable for
            convergence, by default 1e-4.
        ftol : float, optional
            Absolute error in func(xopt) between iterations that is acceptable
            for convergence, by default 1e-4.
        n_restart : int, optional
            The maximal number of random restarts to perform when stagnation is
            detected during the run. The random restart can be switched off by
            setting `n_restart` to zero (the default value).
        verbose : bool, optional
            Verbosity of the output, by default False.
        logger : str, optional
            Name of the logger file, by default None, which turns off the
            logging behaviour.
        random_seed : int, optional
            The seed for pseudo-random number generators, by default None.
        """
        assert isinstance(search_space, RealSpace)
        lb, ub = list(zip(*search_space.bounds))
        self.search_space = search_space
        self.dim: int = search_space.dim
        self.obj_fun: Callable = obj_fun
        self.h: Callable = h
        self.g: Callable = g
        self.minimize: bool = minimize
        self.ftarget: float = ftarget
        self.lb: np.ndarray = set_bounds(lb, self.dim)
        self.ub: np.ndarray = set_bounds(ub, self.dim)
        self.sigma = sigma0
        self.sigma0 = self.sigma
        self.args: Dict = args if args else {}
        self.n_restart: int = max(0, int(n_restart))
        self._restart: bool = False

        self.xopt: np.ndarray = None
        self.fopt: float = None
        self.fopt_penalized: float = None
        self.eval_count: int = 0
        self.iter_count: int = 0
        self.max_FEs: int = int(eval(max_FEs)) if isinstance(max_FEs, str) else max_FEs
        self._better = (lambda a, b: a <= b) if self.minimize else (lambda a, b: a >= b)
        self._init_aux_var(kwargs)
        self._init_covariance(C0)
        self._init_logging_var()

        self.stop_dict: Dict = {}
        self._exception: bool = False
        self.verbose: bool = verbose
        self.logger = logger
        self.random_seed = random_seed

        # parameters for stopping criteria
        # NOTE: `self._delta_f = self.ftol / self._w ** (5 * self.dim)`
        # and `self._w = 0.9` lead to a tolerance of
        # ~`5 * self.dim` iterations of stagnation.
        self.xtol: float = xtol
        self.ftol: float = ftol
        self._w: float = 0.9
        self._delta_x: float = self.xtol / self._w ** (5 * self.dim)
        self._delta_f: float = self.ftol / self._w ** (5 * self.dim)
        self._stop: bool = False
        # set the initial search point
        self.x = x0

    def _init_aux_var(self, opts):
        self.prob_target = opts["p_succ_target"] if "p_succ_target" in opts else 2 / 11
        self.threshold = opts["p_threshold"] if "p_threshold" in opts else 0.44
        self.d = opts["d"] if "d" in opts else 1 + self.dim / 2
        self.ccov = opts["ccov"] if "ccov" in opts else 2 / (self.dim ** 2 + 6)
        self.cp = opts["cp"] if "cp" in opts else 1 / 12
        self.cc = opts["cc"] if "cc" in opts else 2 / (self.dim + 2)

        self.success_rate: float = self.prob_target
        self.pc: np.ndarray = np.zeros(self.dim)
        self._coeff: float = self.cc * (2 - self.cc)

    def _init_covariance(self, C):
        if C is None:
            self._C = np.eye(self.dim)
            self._A = np.eye(self.dim)
        else:
            self.C = C

    def _init_logging_var(self):
        # parameters for logging the history
        self.hist_fopt: List = []
        self.hist_fopt_penalized: List = []
        self.hist_xopt: List = []
        self._hist_delta_x: List = []
        self._hist_delta_f: List = []

    @property
    def random_seed(self):
        return self._random_seed

    @random_seed.setter
    def random_seed(self, seed):
        if seed:
            self._random_seed = int(seed)
            if self._random_seed:
                np.random.seed(self._random_seed)

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        if isinstance(logger, logging.Logger):
            self._logger = logger
            self._logger.propagate = False
            return

        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(logging.DEBUG)
        fmt = LoggerFormatter()

        # create console handler and set level to the vebosity
        SH = list(filter(lambda h: isinstance(h, logging.StreamHandler), self._logger.handlers))
        if self.verbose and len(SH) == 0:
            sh = logging.StreamHandler(sys.stdout)
            sh.setLevel(logging.INFO)
            sh.setFormatter(fmt)
            self._logger.addHandler(sh)

        # create file handler and set level to debug
        # TODO: perhaps also according to the verbosity?
        # TODOL perhaps create a logger class
        FH = list(filter(lambda h: isinstance(h, logging.FileHandler), self._logger.handlers))
        if logger is not None and len(FH) == 0:
            fh = logging.FileHandler(logger)
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(fmt)
            self._logger.addHandler(fh)

        if hasattr(self, "logger"):
            self._logger.propagate = False

    @property
    def C(self):
        return self._C

    @C.setter
    def C(self, C):
        if C is not None:
            try:
                A = np.linalg.cholesky(C)
                if np.all(np.isreal(A)):
                    # TODO: `_A` should be a private attribute
                    self._A = A
                    self._C = C
            except np.linalg.LinAlgError:
                pass

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        if x is not None:
            x = eval(x) if isinstance(x, str) else x
            x = np.asarray(x)
            assert np.all(x - self.lb >= 0)
            assert np.all(x - self.ub <= 0)
        else:
            # sample `x` u.a.r. in `[lb, ub]`
            assert all(~np.isinf(self.lb)) & all(~np.isinf(self.ub))
            x = (self.ub - self.lb) * np.random.rand(self.dim) + self.lb

        self._x = x
        y = self.evaluate(x)
        penalty = self.penalize(x)
        self.tell(x, y, penalty)

    @property
    def sigma(self):
        return self._sigma

    @sigma.setter
    def sigma(self, sigma):
        if sigma is None:
            assert all(~np.isinf(self.lb)) & all(~np.isinf(self.ub))
            sigma = np.max(self.ub - self.lb) / 5
        assert sigma > 0
        self._sigma = sigma

    def run(self):
        while not self._stop:
            self.step()
        return self.xopt, self.fopt, self.stop_dict

    def step(self):
        x = self.ask()
        y = self.evaluate(x)
        self.tell(x, y, self.penalize(x))
        self.logging()
        self.check_stop()
        self.restart()

    def penalize(self, x: np.ndarray):
        """Calculate the dynamic penalty once the constraint functions are provided

        Parameters
        ----------
        x : np.ndarray
            the trial point to check against the constraints
        """
        return dynamic_penalty(x, self.iter_count + 1, self.h, self.g, minimize=self.minimize)

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        self.eval_count += 1
        if isinstance(self.args, (list, tuple)):
            fval = self.obj_fun(x, *self.args)
        elif isinstance(self.args, dict):
            fval = self.obj_fun(x, **self.args)
        return fval

    def restart(self):
        if self._restart:
            self.logger.info("restarting... ")
            self.x = None
            self.sigma = self.sigma0
            self.pc = np.zeros(self.dim)
            self._C = np.eye(self.dim)
            self._A = np.eye(self.dim)
            self._delta_x = self.xtol * 200
            self._delta_f = self.ftol * 200
            self.stop_dict = {}
            self.n_restart -= 1

    def ask(self) -> np.ndarray:
        """The mutation operator

        Parameters
        ----------
        n_point : int, optional
            The number of mutants, which is always 1. This argument is only
            meant to keep the function interface consistant.

        Returns
        -------
        np.ndarray
            The mutation vector
        """
        z = np.random.randn(self.dim).dot(self._A.T)
        x = self._x + self.sigma * z
        x = handle_box_constraint(x, self.lb, self.ub)
        # rounding if a coarser numerical precision is provided
        x = self.search_space.round(x).ravel()
        # NOTE: experimental correction to the step-size when the box constraints are violated
        # self.sigma = np.min(np.abs((x - self._x) / z))
        return x

    def tell(self, x: np.ndarray, y: np.ndarray, penalty: float = 0):
        if self._stop:
            self.logger.info("The optimizer is stopped and `tell` should not be called.")
            return

        # TODO: this might not be uncessary
        if hasattr(y, "__iter__"):
            y = y[0]
        if hasattr(penalty, "__iter__"):
            penalty = penalty[0]
        y_penalized = y + penalty

        if self.xopt is None:
            self.fopt = y
            self.fopt_penalized = y_penalized
            self.xopt = x
            return

        success = self._better(y_penalized, self.fopt_penalized)
        z = (x - self._x) / self._sigma
        self._update_step_size(success)
        self._delta_f *= self._w
        self._delta_x *= self._w

        if success:
            self._delta_f += (1 - self._w) * abs(self.fopt_penalized - y_penalized)
            self._delta_x += (1 - self._w) * np.sqrt(sum((self._x - x) ** 2))
            self.fopt_penalized = y_penalized
            self._x = copy(x)
            self._update_covariance(z)

        if success and penalty == 0:
            self.xopt = copy(self._x)
            self.fopt = y

        self._handle_exception()
        self.iter_count += 1

        if self.verbose:
            self.logger.info(f"iteration {self.iter_count}")
            self.logger.info(f"fopt: {self.fopt}")
            if self.h is not None or self.g is not None:
                _penalty = (self.fopt - self.fopt_penalized) * (-1) ** self.minimize
                self.logger.info(f"penalty: {_penalty[0]:.4e}")
            self.logger.info(f"xopt: {self.xopt.tolist()}")
            self.logger.info(f"sigma: {self._sigma}\n")

    def logging(self):
        self.hist_fopt += [self.fopt]
        self.hist_xopt += [self.xopt.tolist()]
        self.hist_fopt_penalized += [self.fopt_penalized]

    def check_stop(self):
        if self.ftarget is not None and self._better(self.fopt, self.ftarget):
            self.stop_dict["ftarget"] = self.fopt

        if self.eval_count >= self.max_FEs:
            self.stop_dict["FEs"] = self.eval_count

        # TODO: add this as an option: lower and upper bounds for regular sigmas
        if self.sigma < 1e-8 or self.sigma > 1e8:
            self.stop_dict["sigma"] = self.sigma

        if self._delta_f < self.ftol:
            self.stop_dict["ftol"] = self._delta_f

        if self._delta_x < self.xtol:
            self.stop_dict["xtol"] = self._delta_x

        if "ftarget" in self.stop_dict or "FEs" in self.stop_dict:
            self._stop = True
        else:
            if self.n_restart > 0:
                self._restart = bool(self.stop_dict)
            else:
                self._stop = bool(self.stop_dict)

    def _update_covariance(self, z):
        if self.success_rate < self.threshold:
            self.pc = (1 - self.cc) * self.pc + np.sqrt(self._coeff) * z
            self._C = (1 - self.ccov) * self._C + self.ccov * np.outer(self.pc, self.pc)
        else:
            self.pc = (1 - self.cc) * self.pc
            self._C = (1 - self.ccov * (1 - self._coeff)) * self._C + self.ccov * np.outer(
                self.pc, self.pc
            )

        self._C = np.triu(self._C) + np.triu(self._C, 1).T
        self._update_A(self._C)

    def _update_step_size(self, success):
        prob_target = self.prob_target
        self.success_rate = (1 - self.cp) * self.success_rate + self.cp * success
        self._sigma *= np.exp((self.success_rate - prob_target) / (1 - prob_target) / self.d)
        if self._sigma is None:
            breakpoint()

    def _update_A(self, C):
        if np.any(np.isinf(C)):
            self._exception = True
        else:
            try:
                A = np.linalg.cholesky(C)
                if np.any(~np.isreal(A)):
                    self._exception = True
                else:
                    self._A = A
            except np.linalg.LinAlgError:
                self._exception = True

    def _handle_exception(self):
        if self._sigma < 1e-8 or self._sigma > 1e8:
            self._exception = 1

        if self._exception:
            self._C = np.eye(self.dim)
            self.pc = np.zeros(self.dim)
            self._A = np.eye(self.dim)
            self._sigma = self.sigma0
            self._exception = False


class OnePlusOne_Cholesky_CMA(OnePlusOne_CMA):
    """(1+1)-Cholesky-CMA-ES improves its base class algorithm by taking advantage of
    Cholesky's decomposition to update the covariance, which is computationally cheaper

    """

    def _init_covariance(self, C):
        reset = False
        if C is not None:
            try:
                A = np.linalg.cholesky(C)
                if np.any(~np.isreal(A)):
                    reset = True
                else:
                    self.A = A
            except np.linalg.LinAlgError:
                reset = True

        if C is None or reset:
            self.A = np.eye(self.dim)

    @property
    def A(self):
        return self._A

    @A.setter
    def A(self, A):
        assert np.all(np.triu(A, k=1).ravel() == 0)
        self._A = A
        self._A_inv = solve_triangular(A, np.eye(self.dim), lower=True)

    def _update_covariance(self, z):
        cb = self.ccov
        if self.success_rate < self.threshold:
            self.pc = (1 - self.cc) * self.pc + np.sqrt(self._coeff) * z
            ca = 1 - self.ccov
        else:
            self.pc = (1 - self.cc) * self.pc
            ca = (1 - self.ccov) + self.ccov * self.cc * (2 - self.cc)

        w = self.pc.dot(self._A_inv.T)
        w_ = w.dot(self._A_inv)
        L = np.sum(w ** 2)

        self._A += (np.sqrt(1 + L * cb / ca) - 1) / L * np.outer(self.pc, w)
        self._A *= np.sqrt(ca)

        self._A_inv -= (1 - 1 / np.sqrt(1 + L * cb / ca)) / L * np.outer(w, w_)
        self._A_inv *= 1 / np.sqrt(ca)



from __future__ import annotations

import json
from collections import Counter
from copy import copy, deepcopy
from itertools import chain, combinations
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.random import rand
from pyDOE import lhs
from scipy.special import logit
from sobol_seq import i4_sobol_generate

from .._exception import ConstraintEvaluationError
from .node import Node
from .samplers import SCMC
from .variable import Bool, Discrete, Integer, Ordinal, Real, Subset, Variable

__authors__ = "Hao Wang"


class SearchSpace:
    """Search Space Base Class

    Attributes
    ----------
    public:
        dim : int,
            dimensinality of the search space
        bounds : a list of lists,
            each sub-list stores the lower and upper bound for continuous/ordinal variable
            and categorical values for nominal variable
        var_name : a list of str,
            variable names per dimension
        var_type : a list of str,
            variable type per dimension, 'Real': continuous/real-valued,
            'Discrete': discrete, 'Integer': ordinal/integer-valued
        r_mask : a bool array,
            the mask array for continuous variables
        i_mask : a bool array,
            the mask array for integer variables
        d_mask : a bool array,
            the mask array for discrete variables
        id_r : an int array,
            the index array for continuous variables
        id_i : an int array,
            the index array for integer variables
        id_d : an int array,
            the index array for discrete variables

    protected:
        _levels : a dict of lists,
            unique levels for each categorical/discrete variable. The index of
            the corresponding discrete variable serves as the dictionary key.
    """

    _supported_types = (Real, Integer, Ordinal, Discrete, Bool)

    def __init__(self, data: List[Variable], random_seed: int = None, hierarchy: dict = None):
        """Search Space

        Parameters
        ----------
        data : List[Variable]
            a list of variables consistuting the search space
        random_seed : int, optional
            random seed controlling the `sample` function, by default None
        """
        # declarations to fix the pylint warnings..
        self._var_name: List[str] = []
        self._var_type: List[str] = []
        self._bounds: List[tuple] = []
        self._levels: dict = {}

        self.random_seed: int = random_seed
        self._set_data(data)
        SearchSpace.__set_type(self)
        self.structure = hierarchy
        self.__set_conditions()

    @property
    def var_name(self):
        return self._var_name
        # return (self._var_name[0] if self.dim == 1 else self._var_name)

    @var_name.setter
    def var_name(self, var_name):
        if isinstance(var_name, str):
            var_name = [var_name + str(_) for _ in range(self.dim)]
        else:
            assert len(var_name) == self.dim

        for i, name in enumerate(var_name):
            self.data[i].name = name
        self._set_data(self.data)

    @property
    def var_type(self):
        return self._var_type

    @property
    def bounds(self):
        return self._bounds

    @property
    def random_seed(self):
        return self._random_seed

    @random_seed.setter
    def random_seed(self, seed):
        if seed:
            seed = int(seed)
        self._random_seed = seed
        np.random.seed(self._random_seed)

    @staticmethod
    def _ready_args(bounds, var_name, **kwargs):
        """infer the dimension, set the variable name, and ready other arguments"""
        if hasattr(bounds[0], "__iter__") and not isinstance(bounds[0], str):
            bounds = [tuple(b) for b in bounds]
        else:
            bounds = [tuple(bounds)]
        dim = len(bounds)
        out: List[Dict] = [{"bounds": bounds[i]} for i in range(dim)]

        if isinstance(var_name, str):
            if dim > 1:
                var_name = [var_name + str(_) for _ in range(dim)]
            else:
                var_name = [var_name]
        assert len(var_name) == dim
        for i in range(dim):
            out[i]["name"] = var_name[i]

        for key, value in kwargs.items():
            if value is not None:
                if not isinstance(value, (tuple, list)):
                    value = [value] * dim
                assert len(value) == dim
                for i in range(dim):
                    out[i][key] = value[i]
        return out

    def _check_data(self, data):
        assert all([isinstance(d, self._supported_types) for d in data])
        names = np.asarray([var.name for var in data])
        for name, count in Counter(names).items():
            if count > 1:
                idx = np.nonzero(names == name)[0]
                _names = [name + str(i) for i in range(count)]
                for i, k in enumerate(idx):
                    data[k].name = _names[i]

    @property
    def structure(self):
        return self._structure

    @structure.setter
    def structure(self, hierarchy: dict):
        if hierarchy:
            self._structure = Node.from_dict(hierarchy)
        else:
            pass

    @staticmethod
    def __set_type(obj: SearchSpace) -> SearchSpace:
        _type = np.unique(obj.var_type)
        if len(_type) == 1:
            obj.__class__ = eval(_type[0] + "Space")
        else:
            obj.__class__ = SearchSpace
        return obj

    def __set_conditions(self):
        # TODO: perhaps implement a `conditions` property (list of all conditions)
        # TODO: to validate the conditions specified in variables
        # TODO: add conditions when the prefix of some variables appears (conditional parameters)
        for var in self.data:
            if var.conditions is not None:
                pre, _, __ = var.name.rpartition(".")
                if pre:
                    for i, var_name in enumerate(var.var_in_conditions):
                        if not var_name.startswith(pre):
                            var.var_in_conditions[i] = pre + "." + var_name

    def _set_data(self, data):
        """Sanity check on the input data and set the auxiliary variables"""
        self._check_data(data)
        self.data: List = data
        self.dim: int = len(self.data)
        self._bounds = [var.bounds for var in self.data]
        self._var_type = [type(v).__name__ for v in self.data]
        self._var_name = [v.name for v in self.data]
        self._set_index()
        self._set_levels()

    def _set_index(self):
        """set indices for each type of variables"""
        if self.dim > 0:
            for var_type in self._supported_types:
                name = var_type.__name__
                attr_mask = name.lower() + "_mask"
                attr_id = name.lower() + "_id"
                mask = np.asarray(self._var_type) == name
                self.__dict__[attr_mask] = mask
                self.__dict__[attr_id] = np.nonzero(mask)[0]

            self.categorical_id = np.r_[self.discrete_id, self.ordinal_id, self.bool_id]
            self.categorical_mask = np.bitwise_or(
                self.bool_mask, np.bitwise_or(self.discrete_mask, self.ordinal_mask)
            )

    def _set_levels(self):
        # TODO: check if this is still needed
        """Set categorical levels for all nominal variables"""
        if self.dim > 0:
            self.levels = (
                {i: self._bounds[i] for i in self.categorical_id}
                if len(self.categorical_id) > 0
                else {}
            )

    def __getitem__(self, index) -> Union[SearchSpace, Variable]:
        if isinstance(index, (int, slice)):
            data = self.data[index]
        elif hasattr(index, "__iter__") and not isinstance(index, str):
            index = np.array(index)
            if index.dtype.type is np.str_:  # list of variable names
                index = [np.nonzero(np.array(self.var_name) == i)[0][0] for i in index]
            elif index.dtype == bool:  # mask array
                index = np.nonzero(index)[0]
            data = [self.data[i] for i in index]
        elif isinstance(index, str):  # slicing one variable by name
            index = np.nonzero(np.array(self.var_name) == index)[0][0]
            data = self.data[index]
        else:
            raise Exception(f"index type {type(index)} is not supported")

        if isinstance(data, Variable):
            out = data
        elif isinstance(data, list):
            out = SearchSpace(data, self.random_seed)
        return out

    def __setitem__(self, index, value):
        if isinstance(index, (int, slice)):
            self.data[index] = value
        elif isinstance(index, str):
            index = np.nonzero(np.array(self.var_name) == index)[0][0]
            self.data[index] = value
        elif hasattr(index, "__iter__") and not isinstance(index, str):
            if not hasattr(value, "__iter__") or isinstance(value, str):
                value = [value] * len(index)
            for i, v in zip(index, value):
                if isinstance(i, str):
                    k = np.nonzero(np.array(self.var_name) == i)[0][0]
                    self.data[k] = v
                elif isinstance(i, int):
                    self.data[i] = v
                else:
                    raise Exception(f"index type {type(i)} is not supported")
        self._set_data(self.data)

    def __contains__(self, item: Union[str, Variable, SearchSpace, list, dict]) -> bool:
        """check if a name, a variable, a space, or a sample point in the the search space"""
        if isinstance(item, str):
            return item in self.var_name
        if isinstance(item, Variable):
            return item in self.data
        if isinstance(item, SearchSpace):
            return all(map(lambda x: x in self.data, item.data))
        if isinstance(item, list):
            return all([v in self.__getitem__(i) for i, v in enumerate(item)])
        if isinstance(item, dict):
            return all([v in self.__getitem__(i) for i, v in item.items()])
        raise ValueError(f"type {type(item)} is not supported")

    def __len__(self):
        return self.dim

    def __iter__(self) -> Variable:
        i = 0
        while i < self.dim:
            yield self.__getitem__(i)
            i += 1

    def __eq__(self, cs: SearchSpace) -> bool:
        return self.dim == cs.dim and set(self.data) == set(cs.data)

    def __ne__(self, cs: SearchSpace) -> bool:
        return not self.__eq__(cs)

    def __add__(self, space) -> SearchSpace:
        """Direct Sum of two `SearchSpace` instances"""
        assert isinstance(space, SearchSpace)
        # NOTE: the random seed of `self` has the priority
        random_seed = self.random_seed if self.random_seed else space.random_seed
        data = deepcopy(self.data) + space.data
        return SearchSpace(data, random_seed)

    def __radd__(self, space) -> SearchSpace:
        return self.__add__(space)

    def __iadd__(self, space) -> SearchSpace:
        assert isinstance(space, SearchSpace)
        self._set_data(self.data + space.data)
        SearchSpace.__set_type(self)
        return self

    def __sub__(self, space) -> SearchSpace:
        """Substraction of two `SearchSpace` instances"""
        assert isinstance(space, SearchSpace)
        random_seed = self.random_seed if self.random_seed else space.random_seed
        _res = set(self.var_name) - set(space.var_name)
        _index = [self.var_name.index(_) for _ in _res]
        data = [copy(self.data[i]) for i in range(self.dim) if i in _index]
        return SearchSpace(data, random_seed)

    def __rsub__(self, space) -> SearchSpace:
        return self.__sub__(space)

    def __isub__(self, space) -> SearchSpace:
        assert isinstance(space, SearchSpace)
        _res = set(self.var_name) - set(space.var_name)
        _index = [self.var_name.index(_) for _ in _res]
        self._set_data([self.data[i] for i in range(self.dim) if i in _index])
        SearchSpace.__set_type(self)
        return self

    def __mul__(self, N: int) -> SearchSpace:
        """Replicate `self` by copy

        Parameters
        ----------
        N : int
            Replicate `self` N times as a copt

        Returns
        -------
        SearchSpace
            a copy of replicated `self`
        """
        data = [deepcopy(var) for _ in range(max(1, int(N))) for var in self.data]
        obj = SearchSpace(data, self.random_seed)
        obj.__class__ = type(self)
        return obj

    def __rmul__(self, N: int) -> SearchSpace:
        return self.__mul__(N)

    def __imul__(self, N: int) -> SearchSpace:
        """Incrementally replicate

        Parameters
        ----------
        N : int
            Replicate `self` N times

        Returns
        -------
        SearchSpace
            `self`
        """
        self._set_data(
            self.data + [deepcopy(var) for _ in range(max(1, int(N - 1))) for var in self.data]
        )
        return self

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        msg = f"{type(self).__name__} of {self.dim} variables: \n"
        for var in self.data:
            msg += str(var) + "\n"
        return msg

    def filter(self, keys: List[str], invert=False) -> SearchSpace:
        """filter a search space based on a list of variable names

        Parameters
        ----------
        keys : List[str]
            the list of variable names to keep

        Returns
        -------
        Union[Variable, SearchSpace]
            the resulting subspace
        """
        masks = [v in keys for v in self.var_name]
        if invert:
            masks = np.bitwise_not(masks)
        return self[masks]

    @classmethod
    def concat(cls, *args: Tuple[SearchSpace]):
        if len(args) == 1:
            return args[0]

        assert isinstance(args[0], SearchSpace)
        data = list(chain.from_iterable([deepcopy(_.data) for _ in args]))
        return SearchSpace(data)

    def pop(self, index: int = -1) -> Variable:
        value = self.data.pop(index)
        self._set_data(self.data)
        SearchSpace.__set_type(self)
        self.__set_conditions()
        return value

    def remove(self, index: Union[int, str]) -> SearchSpace:
        if isinstance(index, str):
            _index = np.nonzero(np.array(self._var_name) == index)[0]
            if len(_index) == 0:
                raise KeyError(f"The input key {index} not found in `var_name`!")
            _index = _index[0]
        elif hasattr(index, "__iter__"):
            raise KeyError("Multiple indices are not allowed!")
        else:
            _index = index

        self.data.pop(_index)
        self._set_data(self.data)
        self.__set_conditions()
        return SearchSpace.__set_type(self)

    def update(self, space: SearchSpace) -> SearchSpace:
        """Update the search space based on the var_name of the input search space,
        which behaves similarly to the dictionary update. Please note its difference
        to ``self.__add__``

        Parameters
        ----------
        space : SearchSpace
            the input space for the update
        """
        _var_name = space.var_name
        _update = np.array(list(set(_var_name) & set(self.var_name)))
        _add = np.array(list(set(_var_name) - set(self.var_name)))

        _index_update_to = np.nonzero(np.array(self._var_name) == _update)[0]
        _index_update_from = np.nonzero(np.array(_var_name) == _update)[0]
        for i, k in enumerate(_index_update_to):
            j = _index_update_from[i]
            self.data[k] = deepcopy(space.data[j])

        _index_add = np.nonzero(np.array(_var_name) == _add)[0]
        self.data += [space.data[i] for i in _index_add]
        self._set_data(self.data)
        return SearchSpace.__set_type(self)

    def sample(
        self,
        N: int = 1,
        method: str = "uniform",
        h: Callable = None,
        g: Callable = None,
        tol: float = 1e-2,
    ) -> np.ndarray:
        """Sample random points from the search space

        Parameters
        ----------
        N : int, optional
            the number of points to generate, by default 1
        method : str, optional
            the sampling strategy, by default 'uniform'
        h : Callable, optional
            equality constraints, by default None
        g : Callable, optional
            inequality constraints, by default None
        tol : float, optional
            the tolerance on the constraint

        NOTES
        -----
        At this moment, the constraints are handled using the simple Monte Carlo sampling

        Returns
        -------
        np.ndarray
            the sample points in shape `(N, self.dim)`
        """
        # 10 is the minimal number of sample points to take under constraints
        n = max(N, 10) if h or g else N
        constraints = lambda x: np.r_[np.abs(h(x)) if h else [], np.array(g(x)) if g else []]
        S = SCMC(self, constraints, tol=tol).sample(n) if h or g else self._sample(N, method)
        try:
            # NOTE: equality constraints are converted to an epsilon-tude around the
            # corresponding manifold
            idx_h = (
                list(map(lambda x: all(np.isclose(np.abs(h(x)), 0, atol=tol)), S))
                if h
                else [True] * n
            )
            idx_g = list(map(lambda x: np.all(np.asarray(g(x)) <= 0), S)) if g else [True] * n
            idx = np.bitwise_and(idx_h, idx_g)
            S = S[idx, :]
        except Exception as e:
            raise ConstraintEvaluationError(S, str(e)) from None

        # get unique rows
        # S = np.array([list(x) for x in set(tuple(x) for x in S)], dtype=object)
        if len(S) > N:
            S = S[np.random.choice(len(S), N, replace=False)]
        return S

    def _sample(self, N: int = 1, method: str = "uniform") -> np.ndarray:
        # in case this space is empty after slicing
        if self.dim == 0:
            return np.empty(0)

        N = max(int(N), 1)
        X = np.empty((N, self.dim), dtype=object)
        for var_type in self._supported_types:
            attr_id = var_type.__name__.lower() + "_id"
            index = self.__dict__[attr_id]
            if len(index) > 0:  # if such a type of variables exist.
                X[:, index] = getattr(self[index], "_sample")(N, method)
        return X

    def round(self, X: Union[np.ndarray, List[List]]) -> np.ndarray:
        if not isinstance(X, np.ndarray):
            X = np.array(X, dtype=object)
        if len(X.shape) == 1:
            X = X.reshape(1, -1)
        if len(self.real_id) > 0:  # if real-valued variables exist.
            r_subspace = self.__getitem__(self.real_id)
            X[:, self.real_id] = r_subspace.round(X[:, self.real_id].astype(float))
        return X

    def to_linear_scale(self, X: Union[np.ndarray, List[List]]) -> np.ndarray:
        if not isinstance(X, np.ndarray):
            X = np.array(X, dtype=object)
        if len(X.shape) == 1:
            X = X.reshape(1, -1)
        if len(self.real_id) > 0:  # if real-valued variables exist.
            r_subspace = self.__getitem__(self.real_id)
            X[:, self.real_id] = r_subspace.to_linear_scale(X[:, self.real_id].astype(float))
        return X

    def to_dict(self) -> dict:
        out: dict = {}
        for _, var in enumerate(self.data):
            value = {
                "range": var.bounds,
                "type": type(var),
                "N": 1,
            }
            if isinstance(var, Real):
                value["precision"] = var.precision
                value["scale"] = var.scale
            elif isinstance(var, Integer):
                value["step"] = var.step

            if hasattr(var, "conditions"):
                value["conditions"] = var.conditions
                # TODO: also export var._action?
            out[var.name] = value
        return out

    def to_json(self, file: str):
        with open(file, "w") as f:
            json.dump(self.to_dict(), f)

    @classmethod
    def from_dict(cls, param: dict) -> SearchSpace:
        """Create a search space object from input dictionary

        Parameters
        ----------
        param : dict
            A dictionary that describes the search space
        space_name : bool, optional
            Whether a (multi-dimensional) subspace should be named. If this named space
            is a subspace a whole search space, for a solution sampled from the whole space, its
            components pertaining to this subspace will be grouped together under the key
            `space_name`, when this solution is converted to a dictionary/json
            (see `SearchSpace.to_dict`).
        source : string, optional
            Where the dictionary originates from. Can be either 'default' or 'irace'

        Returns
        -------
        SearchSpace
        """
        # TODO: implement `grouping`
        assert isinstance(param, dict)

        variables = []
        for k, v in param.items():
            if "range" in v:
                bounds = v["range"]
                if not hasattr(bounds[0], "__iter__") or isinstance(bounds[0], str):
                    bounds = tuple(bounds)
            else:
                bounds = ()

            N = range(int(v["N"])) if "N" in v else range(1)
            default_value = v["defualt"] if "default" in v else None
            if v["type"] in ["r", "real"]:  # real-valued parameter
                precision = v["precision"] if "precision" in v else None
                scale = v["scale"] if "scale" in v else "linear"
                _vars = [
                    Real(
                        bounds,
                        name=k,
                        default_value=default_value,
                        precision=precision,
                        scale=scale,
                    )
                    for _ in N
                ]
            elif v["type"] in ["i", "int", "integer"]:  # integer-valued parameter
                _vars = [
                    Integer(bounds, name=k, default_value=default_value, step=v.pop("step", 1))
                    for _ in N
                ]
            elif v["type"] in ["o", "ordinal"]:  # ordinal parameter
                _vars = [Ordinal(bounds, name=k, default_value=default_value) for _ in N]
            elif v["type"] in ["c", "cat"]:  # category-valued parameter
                _vars = [Discrete(bounds, name=k, default_value=default_value) for _ in N]
            elif v["type"] in ["s", "subset"]:  # subset parameter
                _vars = [Subset(bounds, name=k, default_value=default_value) for _ in N]
            elif v["type"] in ["b", "bool"]:  # Boolean-valued
                _vars = [Bool(name=k, default_value=default_value) for _ in N]
            variables += _vars
        return SearchSpace(variables)

    @classmethod
    def from_json(cls, file: str) -> SearchSpace:
        """Create a seach space from a json file

        Parameters
        ----------
        file : str
            Path to the input json file

        Returns
        -------
        SearchSpace
            an `SearchSpace` object converted from the json file
        """
        with open(file, "r") as f:
            return cls.from_dict(json.load(f))

    def pprint(self):
        print(str(self))
        if self.structure:
            for root in self.structure:
                root.pprint()


class RealSpace(SearchSpace):
    """Space of real values"""

    def __init__(
        self,
        bounds: List,
        var_name: Union[str, List[str]] = "real",
        default_value: Union[float, List[float]] = None,
        precision: Union[int, List[int]] = None,
        scale: Union[str, List[str]] = None,
        **kwargs,
    ):
        out = self._ready_args(
            bounds, var_name, default_value=default_value, precision=precision, scale=scale
        )
        data = [Real(**_) for _ in out]
        super().__init__(data, **kwargs)

    def _sample(self, N: int = 1, method: str = "uniform") -> np.ndarray:
        bounds = np.array([var._bounds_transformed for var in self.data])
        lb, ub = bounds[:, 0], bounds[:, 1]
        if method == "uniform":  # uniform random samples
            X = (ub - lb) * rand(N, self.dim) + lb
        elif method == "LHS":  # Latin hypercube sampling
            if N == 1:
                X = (ub - lb) * rand(N, self.dim) + lb
            else:
                X = (ub - lb) * lhs(self.dim, samples=N, criterion="maximin") + lb
        elif method == "sobol":
            X = (ub - lb) * i4_sobol_generate(self.dim, N) + lb
        return self.round(self.to_linear_scale(X))

    def round(self, X: Union[np.ndarray, List[List]]) -> np.ndarray:
        X = np.atleast_2d(X).astype(float)
        assert X.shape[1] == self.dim
        for i, var in enumerate(self.data):
            X[:, i] = var.round(X[:, i])
        return X

    def to_linear_scale(self, X: Union[np.ndarray, List[List]]) -> np.ndarray:
        X = np.atleast_2d(X).astype(float)
        assert X.shape[1] == self.dim
        for i, var in enumerate(self.data):
            X[:, i] = var.to_linear_scale(X[:, i])
        return X


class _DiscreteSpace(SearchSpace):
    """Space of discrete values"""

    def round(self, X: Union[np.ndarray, List[List]]) -> np.ndarray:
        """do nothing since this method is not valid for this class"""
        return X

    def to_linear_scale(self, X: Union[np.ndarray, List[List]]) -> np.ndarray:
        """do nothing since this method is not valid for this class"""
        return X

    def _sample(self, N: int = 1, method: str = "uniform") -> np.ndarray:
        if isinstance(self, IntegerSpace):
            dtype = int
        elif isinstance(self, BoolSpace):
            dtype = bool
        else:
            dtype = object

        X = np.empty((N, self.dim), dtype=dtype)
        for i in range(self.dim):
            X[:, i] = self.data[i].sample(N, method=method)
        return X


class SubsetSpace(_DiscreteSpace):
    """A discrete space created by enumerating all subsets of the input `bounds`"""

    def __init__(
        self,
        bounds: List,
        var_name: Union[str, List[str]] = "subset",
        default_value: Union[int, List[int]] = None,
        **kwargs,
    ):
        out = self._ready_args(bounds, var_name, default_value=default_value)
        data = [Subset(**_) for _ in out]
        super().__init__(data, **kwargs)


class IntegerSpace(_DiscreteSpace):
    """Space of contiguous integer values"""

    def __init__(
        self,
        bounds: List,
        var_name: Union[str, List[str]] = "integer",
        default_value: Union[int, List[int]] = None,
        step: Optional[Union[int, float]] = 1,
        **kwargs,
    ):
        out = self._ready_args(bounds, var_name, default_value=default_value, step=step)
        data = [Integer(**_) for _ in out]
        super().__init__(data, **kwargs)


class OrdinalSpace(_DiscreteSpace):
    """Space of ordinal values"""

    def __init__(
        self,
        bounds: List[str, int, float],
        var_name: Union[str, List[str]] = "ordinal",
        default_value=None,
        **kwargs,
    ):
        out = self._ready_args(bounds, var_name, default_value=default_value)
        data = [Ordinal(**_) for _ in out]
        super().__init__(data, **kwargs)


class DiscreteSpace(_DiscreteSpace):
    """Space of discrete values"""

    def __init__(
        self,
        bounds: List[str, int, float],
        var_name: Union[str, List[str]] = "discrete",
        default_value=None,
        **kwargs,
    ):
        out = self._ready_args(bounds, var_name, default_value=default_value)
        data = [Discrete(**_) for _ in out]
        super().__init__(data, **kwargs)


class BoolSpace(_DiscreteSpace):
    """Space of Bool values"""

    def __init__(
        self,
        var_name: Union[str, List[str]] = "bool",
        default_value: Union[bool, List[bool]] = None,
        **kwargs,
    ):
        out = self._ready_args((False, True), var_name, default_value=default_value)
        data = [Bool(**_) for _ in out]
        super().__init__(data, **kwargs)



import sys

import numpy as np

sys.path.insert(0, "../")
from bayes_optim.acquisition_optim import MIES, OnePlusOne_Cholesky_CMA
from bayes_optim.search_space import Bool, Discrete, Integer, Ordinal, Real, RealSpace, SearchSpace


def obj_fun_min(x):
    return np.sum(x ** 2)


def obj_fun_max(x):
    return -np.sum(x ** 2)


def obj_fun_mies(x):
    x1, x2, x3, x4, x5 = x.tolist()
    return x1 ** 2 + abs(x5) + x4 - 2 + 10 * int(x3 == "B") + 5 * int(not x2)


def h(x):
    return np.sum(x) - 1


def g(x):
    return 1 - np.sum(x)


def test_OnePlusOne_Cholesky_CMA():
    _, __, stop_dict = OnePlusOne_Cholesky_CMA(
        search_space=RealSpace([-5, 5]) * 2,
        obj_fun=obj_fun_min,
        sigma0=2,
        ftol=1e-2,
        verbose=False,
        random_seed=42,
    ).run()
    assert "ftol" in stop_dict


def test_OnePlusOne_Cholesky_CMA_constraint():
    xopt, _, __, = OnePlusOne_Cholesky_CMA(
        search_space=RealSpace([-5, 5]) * 2,
        obj_fun=obj_fun_max,
        h=h,
        sigma0=2,
        max_FEs=500,
        minimize=False,
        verbose=False,
        random_seed=42,
    ).run()
    assert np.isclose(h(xopt), 0, atol=1e-1)

    OnePlusOne_Cholesky_CMA(
        search_space=RealSpace([-5, 5]) * 2,
        obj_fun=obj_fun_min,
        g=g,
        sigma0=2,
        max_FEs=1000,
        verbose=False,
        random_seed=42,
    ).run()
    assert np.isclose(g(xopt), 0, atol=1e-1)


def test_MIES():
    space = SearchSpace(
        [
            Real([0, 1], "x1"),
            Bool("x2"),
            Discrete(["A", "B", "C"], "x3"),
            Ordinal([0, 2, 4, 6, 8], "x4"),
            Integer([0, 10], "x5"),
        ]
    )
    _, __, stop_dict = MIES(
        search_space=space,
        max_eval=100,
        obj_func=obj_fun_mies,
        verbose=False,
        eval_type="list",
        minimize=True,
    ).optimize()
    assert "max_eval" in stop_dict



# -*- coding: utf-8 -*-
"""
Created on Tue Jan 29 11:26:41 2013

@author: wangronin
"""

import numpy as np
from numpy import (
    add,
    append,
    arange,
    argsort,
    array,
    atleast_2d,
    ceil,
    diag,
    dot,
    exp,
    eye,
    floor,
    inf,
    inner,
    isinf,
    isreal,
    linspace,
    log,
    mod,
    newaxis,
    ones,
    outer,
    power,
    r_,
    real,
    size,
    sqrt,
    triu,
    zeros,
)
from numpy.linalg import LinAlgError, cond, eigh, qr
from numpy.random import rand, randn, shuffle
from scipy.stats import chi

# import hello as h
from .boundary_handling import boundary_handling

# My fast routines...
abs = np.abs
max = np.max
min = np.min
sum = np.sum
norm = np.linalg.norm
any = np.any
all = np.all

# TODO: modularize this function, and possiblly fit it into a optimizer class
class cma_es(object):
    """

    My toy CMA-ES... with lots of variants of mutation operators...
    TODO: complete Python __doc__ of this function
    """

    def __init__(
        self,
        dim,
        init_wcm,
        fitnessfunc,
        opts,
        sampling_method=0,
        is_register=False,
        is_minimize=True,
        restart="IPOP",
    ):

        self.stop_dict = {}
        self.offspring = None
        self.sel = None
        self.z = None
        self.evalcount = 0
        self.eigeneval = 0
        self.fitnessfunc = fitnessfunc if is_minimize else lambda x: -fitnessfunc(x)
        self.fitness = None
        self.fitness_rank = None
        self.sampling_method = sampling_method
        self.is_minimize = is_minimize

        # TODO: implement BIPOP
        if restart == "IPOP":
            self.restart = restart
            self.restart_count = 0
            self.inc_popsize = 2
            self.restart_budget = opts["restart_budget"] if "restart_budget" in opts else int(20)

        # Initialize internal strategy parameters
        self.wcm = eval(init_wcm) if isinstance(init_wcm, str) else init_wcm
        self.lb = eval(opts["lb"]) if isinstance(opts["lb"], str) else opts["lb"]
        self.ub = eval(opts["ub"]) if isinstance(opts["ub"], str) else opts["ub"]
        self.lb = atleast_2d(self.lb)
        self.ub = atleast_2d(self.ub)

        if self.lb.shape[1] != 1:
            self.lb = self.lb.T
        if self.ub.shape[1] != 1:
            self.ub = self.ub.T

        self.eval_budget = (
            int(eval(opts["eval_budget"]))
            if isinstance(opts["eval_budget"], str)
            else int(opts["eval_budget"])
        )

        self.wcm = self.wcm.reshape(-1, 1)
        self.dim = dim
        self.sigma0 = opts["sigma_init"]
        self.sigma = self.sigma0
        self.f_target = opts["f_target"] if self.is_minimize else -opts["f_target"]

        # Strategy parameters: Selection
        self._lambda = opts["_lambda"] if "_lambda" in opts else int(4 + floor(3 * log(dim)))
        if isinstance(self._lambda, str):
            self._lambda = eval(self._lambda)
        _mu_prime = (self._lambda - 1) / 2.0
        self._mu = opts["_mu"] if "_mu" in opts else int(ceil(_mu_prime))
        if isinstance(self._mu, str):
            self._mu = eval(self._mu)

        # TODO : new weight setting weighted recombination
        self.weights = log(_mu_prime + 1.0) - log(arange(1, self._mu + 1)[:, newaxis])
        self.weights = self.weights / sum(self.weights)
        self.mueff = sum(self.weights) ** 2.0 / sum(self.weights ** 2)

        self.wcm_old = self.wcm
        self.xopt = self.wcm
        self.fopt = np.inf

        self.pc = zeros((dim, 1))
        self.ps = zeros((dim, 1))
        self.e_vector, self.e_value = eye(dim), ones((dim, 1))
        self.C = dot(self.e_vector, self.e_value * self.e_vector.T)
        self.invsqrt_C = dot(self.e_vector, self.e_value ** -1.0 * self.e_vector.T)

        # Strategy parameter: Adaptation
        # TODO: verify and update the strategy parameters
        self.cc = (4.0 + self.mueff / self.dim) / (self.dim + 4.0 + 2.0 * self.mueff / self.dim)
        self.cs = (self.mueff + 2.0) / (self.dim + self.mueff + 5.0)
        if self._mu == 1:
            self.c_1 = min([2, self._lambda / 3.0]) / ((self.dim + 1.3) ** 2 + self.mueff)
            self.damps = 0.3 + 2.0 * self.mueff / self._lambda + self.cs

        else:  # Original settings
            self.c_1 = 2 / ((self.dim + 1.3) ** 2 + self.mueff)
            self.damps = (
                1.0 + 2 * np.max([0, sqrt((self.mueff - 1) / (self.dim + 1)) - 1]) + self.cs
            )
        self.c_mu = min(
            [
                1 - self.c_1,
                2 * (self.mueff - 2.0 + 1.0 / self.mueff) / ((self.dim + 2) ** 2 + self.mueff),
            ]
        )
        # TODO: verify this
        #    c_mu = 2 * (mueff-2.+1/mueff) / ((dim+2)**2+mueff)
        #    cc = 4.0 / (dim + 4.0)
        #    cs = (mueff+2.) / (dim+mueff+3.)

        # damps parameter tuning
        if "damps" in opts:
            self.damps = opts["damps"]
        else:
            # TODO: Parameter setting for mirrored orthogonal sampling
            # damps tuning for mirrored orthogonal sampling
            if self.sampling_method == 1:
                self.damps = 1.032 - 0.7821 * self.mueff / self._lambda + self.cs

            if self.sampling_method == 4 or self.sampling_method == 8:
                if 1 < 2:
                    # damps setting for small _lambda
                    self.damps = (
                        -0.6314 * (sqrt((self.mueff + 0.1572) / (self.dim + 1.647)) + 0.869)
                        + self.cs
                        + 1.49
                    )
                else:
                    # damps setting for large _lambda
                    self.damps = 1.17 + 4.625 * self.mueff / self._lambda - 2.704 * self.cs

            # Optimal damps setting for mirrored orthogonal sampling
            if self.sampling_method == 7:
                self.damps *= 0.3

        # Axuilirary variables
        self.chiN = dim ** 0.5 * (1 - 1.0 / (4 * dim) + 1.0 / (21 * dim ** 2))
        self.aux = array([])

        # rescaling constant for derandomized step-size
        self.scale = self.chiN

        # Parameters for restart heuristics and warnings
        self.is_stop_on_warning = restart == "IPOP" or restart == "BIPOP"
        self.flg_warning = 0

        self.tolx = 1e-12 * self.sigma
        self.tolupx = 1e3 * max(self.sigma)
        self.tolfun = 1e-12
        self.nbin = int(10 + ceil(30.0 * self.dim / self._lambda))
        self.histfunval = zeros(self.nbin)

        # evolution history registration
        self.is_info_register = is_register
        if is_register:
            self.histsigma = zeros(self.eval_budget)
            self.hist_condition_number = zeros(self.eval_budget)
            self.hist_e_value = zeros((dim, self.eval_budget))
            self.hist_fbest = zeros(self.eval_budget)
            self.hist_xbest = zeros((self.eval_budget, self.dim))

            start = 200
            self.histindex = list(r_[0, linspace(start, self.eval_budget, 10)])
            self.histdist = zeros(size(self.histindex))
            self.ii = 0

        # if performaning pairwise selection
        self.is_pairwise_selection = self._mu != 1 and (
            self.sampling_method == 11
            or self.sampling_method == 4
            or self.sampling_method == 8
            or self.sampling_method == 1
            or self.sampling_method == 7
            or self.sampling_method == 44
        )

    def mutation(self):
        # ---------------------------- Mutation --------------------------------

        # Mirroring
        mode = self.sampling_method
        dim, _lambda, sigma, evalcount, scale, aux = (
            self.dim,
            self._lambda,
            self.sigma,
            self.evalcount,
            self.scale,
            self.aux,
        )

        if mode == 1 or mode == 11:
            if mod(evalcount + _lambda, 2) != 0:
                half = int(ceil(_lambda / 2.0))
                z = randn(dim, half)
                aux = -z[:, -1].reshape(-1, 1)
                z = append(z, -z[:, :-1], axis=1)
            else:
                half = int(floor(_lambda / 2.0))
                z = randn(dim, half)
                z = append(z, -z, axis=1)
                if len(aux) != 0:
                    z = append(aux, z, axis=1)
            self.half = half
        # Derandomized step size
        elif mode == 3:
            z = randn(dim, _lambda)
            z = scale * z / sqrt(sum(power(z, 2), 0))

        # Orthogonal mirrored sampling
        elif mode == 4 or mode == 7 or mode == 44:
            if mod(evalcount + _lambda, 2) != 0:
                half = int(ceil(_lambda / 2.0))
                z = zeros((dim, half))
                n = int(min([dim, half]))
                if dim < half:
                    z[:, dim:] = randn(dim, half - dim)
                q = qr(randn(dim, n))[0]
                l = chi.rvs(dim, size=n)
                z[:, 0:n] = l * q
                aux = -z[:, -1].reshape(-1, 1)
                z = append(z, -z[:, :-1], axis=1)
            else:
                half = int(floor(_lambda / 2.0))
                z = zeros((dim, half))
                n = int(min([dim, half]))
                if dim < half:
                    z[:, dim:] = randn(dim, half - dim)
                q = qr(randn(dim, n))[0]
                l = chi.rvs(dim, size=n)
                z[:, 0:n] = l * q
                z = append(z, -z, axis=1)
                if len(aux) != 0:
                    z = append(aux, z, axis=1)
            self.half = half
        # Orthogonal mirrored sampling...well
        elif mode == 8:
            if mod(evalcount + _lambda, 2) != 0:
                half = ceil(_lambda / 2.0)
                z = zeros((dim, half))
                n = min([dim, half])
                if dim < half:
                    z[:, dim:] = randn(dim, half - dim)
                tmp = randn(dim, n)
                l = sqrt(np.sum(power(tmp, 2), 0))
                q = qr(tmp)[0]
                z[:, 0:n] = l * q
                aux = -z[:, -1].reshape(-1, 1)
                z = append(z, -z[:, :-1], axis=1)
            else:
                half = floor(_lambda / 2.0)
                z = zeros((dim, half))
                n = min([dim, half])
                if dim < half:
                    z[:, dim:] = randn(dim, half - dim)
                tmp = randn(dim, n)
                l = sqrt(np.sum(power(tmp, 2), 0))
                q = qr(tmp)[0]
                z[:, 0:n] = l * q
                z = append(z, -z, axis=1)
                if len(aux) != 0:
                    z = append(aux, z, axis=1)
            self.half = half
        # Orthogonal sampling (random rotation)
        elif mode == 5:
            pass
            # z = dot(rand_orth_mat(dim), eye(dim))
            # n = dim
            # if dim > _lambda:
            #     p = arange(0, dim)
            #     shuffle(p)
            #     z = z[:, p[0:_lambda]]
            #     n = _lambda
            # l = chi.rvs(dim, size=n)
            # sign = rand(n)
            # sign[sign > .5] = 1
            # sign[sign <= .5] = -1
            # z = sign * l * z
            # if dim < _lambda:
            #     ss = randn(dim, _lambda-dim)
            #     z = append(z, ss, axis=1)

        # Orthogonal sampling (Gram-Schmidt)
        elif mode == 6:
            z = zeros((dim, _lambda))
            n = min([dim, _lambda])
            if dim < _lambda:
                z[:, dim:] = randn(dim, _lambda - dim)
            q = qr(randn(dim, n))[0]
            l = chi.rvs(dim, size=n)

        elif mode == 9:
            if mod(evalcount + _lambda, 2) != 0:
                half = int(ceil(_lambda / 2.0))
                z = randn(dim, half)
                q = randn(dim, half)
                p = array(
                    [
                        q[:, i] - inner(q[:, i], z[:, i]) * z[:, i] / norm(z[:, i]) ** 2
                        for i in range(half)
                    ]
                ).T
                p = (p / sqrt(sum(power(p, 2), 0))) * sqrt(sum(power(q, 2), 0))
                aux = p[:, -1].reshape(-1, 1)
                z = append(z, p[:, :-1], axis=1)
            else:
                half = int(floor(_lambda / 2.0))
                z = randn(dim, half)
                q = randn(dim, half)
                p = array(
                    [
                        q[:, i] - inner(q[:, i], z[:, i]) * z[:, i] / norm(z[:, i]) ** 2
                        for i in range(half)
                    ]
                ).T
                p = (p / sqrt(sum(power(p, 2), 0))) * sqrt(sum(power(q, 2), 0))
                z = append(z, p, axis=1)
                if len(aux) != 0:
                    z = append(aux, z, axis=1)
            self.half = half
        # Standard mutation
        else:
            z = randn(dim, _lambda)

        self.z = z
        self.offspring = add(self.wcm, sigma * self.e_vector.dot(self.e_value * self.z))

    def constraint_handling(self):
        self.offspring = boundary_handling(self.offspring, self.lb, self.ub)

    def evaluation(self):
        try:
            self.fitness = self.fitnessfunc(self.offspring)
        except Exception:  # if the fitness function evaluates a single point each time
            self.fitness = np.array([self.fitnessfunc(_) for _ in self.offspring.T])
        self.evalcount += self._lambda
        self.fitness_rank = argsort(self.fitness)
        self.fitness_true = self.fitness * (-1) ** (~self.is_minimize)

    def update(self):
        # Cumulation: Update evolution paths
        cc, cs, c_1, c_mu = self.cc, self.cs, self.c_1, self.c_mu
        wcm, wcm_old, mueff, invsqrt_C = self.wcm, self.wcm_old, self.mueff, self.invsqrt_C
        evalcount, _lambda = self.evalcount, self._lambda

        self.ps = (1 - cs) * self.ps + sqrt(cs * (2 - cs) * mueff) * dot(
            invsqrt_C, (wcm - wcm_old) / self.sigma
        )
        hsig = sum(self.ps ** 2.0) / (
            1.0 - (1.0 - cs) ** (2.0 * evalcount / _lambda)
        ) / self.dim < 2.0 + 4.0 / (self.dim + 1.0)
        self.pc = (1 - cc) * self.pc + hsig * sqrt(cc * (2.0 - cc) * mueff) * (
            wcm - wcm_old
        ) / self.sigma

        offset = (self.offspring[:, self.sel] - wcm_old) / self.sigma
        self.C = (
            (1.0 - c_1 - c_mu) * self.C
            + c_1 * (outer(self.pc, self.pc) + (1.0 - hsig) * cc * (2 - cc) * self.C)
            + c_mu * dot(offset, self.weights * offset.T)
        )
        # Adapt step size sigma
        self.sigma = self.sigma * exp((norm(self.ps) / self.chiN - 1) * self.cs / self.damps)

        if 11 < 3:  # TODO: Saw in the Hansen's codde, reason for this...
            self.sigma = self.sigma * exp(
                min([1, (norm(self.ps) / self.chiN - 1) * self.cs / self.damps])
            )

    def updateBD(self):
        # Eigen decomposition
        C = self.C  # lastest setting for
        C = triu(C) + triu(C, 1).T  # eigen decomposition
        if any(isinf(C)) > 1:  # interval
            self.flg_warning ^= 2 ** 0
        else:
            try:
                w, e_vector = eigh(C)
                e_value = sqrt(list(map(complex, w))).reshape(-1, 1)
                if any(~isreal(e_value)) or any(isinf(e_value)):
                    if self.is_stop_on_warning:
                        self.stop_dict["EigenvalueError"] = True
                    else:
                        self.flg_warning ^= 2 ** 1
                else:
                    self.e_value = real(e_value)
                    self.e_vector = e_vector
                    self.invsqrt_C = dot(e_vector, e_value ** -1 * e_vector.T)
            except LinAlgError:
                if self.is_stop_on_warning:
                    self.stop_dict["linalgerror"] = True
                else:
                    self.flg_warning ^= 2 ** 1

    def info_register(self):
        # Register historical info
        evalcount, _lambda, fitness, offspring = (
            self.evalcount,
            self._lambda,
            self.fitness,
            self.offspring,
        )
        if self.is_info_register:
            if self.ii < len(self.histindex):
                if abs(self.evalcount - self.histindex[self.ii]) <= self._lambda:
                    self.histdist[self.ii] = norm(self.wcm_old)
                    self.ii += 1
            self.histsigma[evalcount - _lambda : evalcount] = self.sigma
            self.hist_condition_number[evalcount - _lambda : evalcount] = (
                max(self.e_value) ** 2 / min(self.e_value) ** 2.0
            )
            self.hist_e_value[:, evalcount - _lambda : evalcount] = self.e_value
            self.hist_xbest[evalcount - _lambda : evalcount, :] = offspring[:, self.sel[0]]
            self.hist_fbest[evalcount - _lambda : evalcount] = fitness[self.sel[0]]

    def check_stop_criteria(self):
        # -------------------------- Restart criterion ------------------------------
        is_stop_on_warning = self.is_stop_on_warning
        sigma, evalcount, _lambda, fitness = self.sigma, self.evalcount, self._lambda, self.fitness

        self.stop_dict["ftarget"] = True if self.fopt <= self.f_target else False
        self.stop_dict["maxfevals"] = True if self.evalcount >= self.eval_budget else False

        if self.evalcount != 0:
            if np.any(fitness == inf) or np.any(fitness == np.nan):
                # TODO: nasty error to be debugged
                raise Exception("Somthing is wrong!")

            if (sigma < 1e-16) or (sigma > 1e6):
                self.flg_warning = True

            diagC = diag(self.C).reshape(-1, 1)
            self.histfunval[int(mod(evalcount / _lambda - 1, self.nbin))] = fitness[self.sel[0]]
            if (
                mod(evalcount, _lambda) == self.nbin
                and max(self.histfunval) - min(self.histfunval) < self.tolfun
            ):
                if is_stop_on_warning:
                    self.stop_dict["tolfun"] = True
                else:
                    self.flg_warning = True

            # Condition covariance
            if cond(self.C) > 1e14:
                if is_stop_on_warning:
                    self.stop_dict["conditioncov"] = True
                else:
                    self.flg_warning = True

            # TolX
            tmp = append(abs(self.pc), sqrt(diagC), axis=1)
            if all(self.sigma * (max(tmp, axis=1)) < self.tolx):
                if is_stop_on_warning:
                    self.stop_dict["TolX"] = True
                else:
                    self.flg_warning = True

            # TolUPX
            if any(sigma * sqrt(diagC)) > self.tolupx:
                if is_stop_on_warning:
                    self.stop_dict["TolUPX"] = True
                else:
                    self.flg_warning = True

            # No effective axis
            a = int(mod(evalcount / _lambda - 1, self.dim))
            if all(0.1 * sigma * self.e_value[a, 0] * self.e_vector[:, a] + self.wcm == self.wcm):
                if is_stop_on_warning:
                    self.stop_dict["noeffectaxis"] = True
                else:
                    sigma *= exp(0.2 + self.cs / self.damps)

            # No effective coordinate
            if any(0.2 * sigma * sqrt(diagC) + self.wcm == self.wcm):
                if is_stop_on_warning:
                    self.stop_dict["noeffectcoord"] = True
                else:
                    self.C += (self.c_1 + self.c_mu) * diag(
                        diagC * (self.wcm == self.wcm + 0.2 * sigma * sqrt(diagC))
                    )
                    sigma *= exp(0.05 + self.cs / self.damps)
            # Adjust step size in case of equal function values
            if (
                fitness[self.sel[0]]
                == fitness[self.sel[int(min([ceil(0.1 + _lambda / 4.0), self._mu - 1]))]]
            ):
                if is_stop_on_warning:
                    self.stop_dict["flatfitness"] = True
                else:
                    sigma *= exp(0.2 + self.cs / self.damps)

        # Handling warnings: Internally rectification of strategy paramters
        if self.flg_warning != 0:
            self.reset_state()
            self.flg_warning = False

    def reset_state(self):
        self.C = eye(self.dim)
        self.e_vector = eye(self.dim)
        self.e_value = ones((self.dim, 1))
        self.invsqrt_C = eye(self.dim)
        self.pc = zeros((self.dim, 1))
        self.ps = zeros((self.dim, 1))
        self.sigma = self.sigma0

    def reset_stop_dict(self):
        for key, _ in self.stop_dict.iteritems():
            self.stop_dict[key] = False

    def optimize(self):
        # TODO: use IPOP for now, to implement BIPOP method
        while self.restart_count < self.restart_budget:
            while True:
                self.info_register()
                self.mutation()
                self.constraint_handling()
                self.evaluation()

                # --------------------------- Comma selection ----------------------------
                # pairwise selection for mirroring
                if self.is_pairwise_selection:
                    #                    self.sel = pairwise_selection(self.fitness_rank, self.half, self._mu)
                    pass
                else:
                    self.sel = self.fitness_rank[0 : self._mu]

                # ------------------------- Weighted recombination ----------------------
                self.wcm_old = self.wcm
                self.wcm = dot(self.offspring[:, self.sel], self.weights)
                self.update()

                # update the eigenvectors and eigenvalues. for computational time concern
                if (
                    self.evalcount - self.eigeneval
                    > self._lambda / (self.c_1 + self.c_mu) / self.dim / 10
                ):
                    self.eigeneval = self.evalcount
                    self.updateBD()

                if self.fopt > self.fitness[self.sel[0]]:
                    self.fopt = self.fitness[self.sel[0]]
                    self.xopt = self.offspring[:, self.sel[0]].reshape(self.dim, -1)

                self.check_stop_criteria()

                if any(array(self.stop_dict.values()) == True):
                    break

            # if the termination criteria are met...
            if self.stop_dict["ftarget"] or self.stop_dict["maxfevals"]:
                break

            self.restart_count += 1
            self._lambda *= self.inc_popsize
            _mu_prime = (self._lambda - 1) / 2.0
            self._mu = int(ceil(_mu_prime))
            self.weights = log(_mu_prime + 1.0) - log(arange(1, self._mu + 1)[:, newaxis])
            self.weights = self.weights / sum(self.weights)
            self.mueff = sum(self.weights) ** 2.0 / sum(self.weights ** 2)

            self.wcm = rand(self.dim, 1) * (self.ub - self.lb) + self.lb
            self.reset_state()
            self.reset_stop_dict()

        return self.xopt, self.fopt, self.evalcount, self.stop_dict



from typing import Callable, Dict

# default aquisition optimization budget
default_AQ_max_FEs: Dict = {
    "MIES": lambda dim: int(1e3 * dim),
    "BFGS": lambda dim: int(1e2 * dim),
    "OnePlusOne_Cholesky_CMA": lambda dim: int(1e3 * dim),
}

default_AQ_n_restart: Callable = lambda dim: int(5 * dim)
default_AQ_wait_iter: int = 3



import os
import sys
from time import time

import cocoex
import cocopp
import numpy as np

sys.path.insert(0, "../")
from bayes_optim import BO, RealSpace
from bayes_optim.surrogate import GaussianProcess, trend


def run_optimizer(optimizer, dim, fID, instance, logfile, lb, ub, max_FEs, data_path, bbob_opt):
    """Parallel BBOB/COCO experiment wrapper"""
    # Set different seed for different processes
    start = time()
    seed = np.mod(int(start) + os.getpid(), 1000)
    np.random.seed(seed)

    data_path = os.path.join(data_path, str(instance))
    max_FEs = eval(max_FEs)

    suite_filter_options = "dimensions: %d instance_indices: %d year:2019" % (dim, instance)
    suite = cocoex.Suite("bbob", "", suite_filter_options)

    observer = cocoex.Observer("bbob", "result_folder: " + data_path)
    f = suite.get_problem(fID)
    f.observe_with(observer)
    assert dim == f.dimension

    opt = optimizer(dim, f, -np.inf, max_FEs, f.lower_bounds, f.upper_bounds, logfile)
    opt.run()

    f.finalizerun()
    with open("out", "a") as fout:
        fout.write(
            "{} on f{} in {}D, instance {}: FEs={}, fbest-ftarget={:.4e}, "
            "elapsed time [m]: {:.3f}\n".format(
                optimizer,
                fID,
                dim,
                instance,
                f.evaluations,
                f.fbest - f.ftarget,
                (time() - start) / 60.0,
            )
        )


def test_BO(dim, obj_fun, ftarget, max_FEs, lb, ub, logfile):

    space = RealSpace(list(zip(lb, ub)))
    mean = trend.constant_trend(dim, beta=None)  # equivalent to Ordinary Kriging
    thetaL = 1e-10 * (ub - lb) * np.ones(dim)
    thetaU = 10 * (ub - lb) * np.ones(dim)
    theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

    model = GaussianProcess(
        mean=mean,
        corr="matern",
        theta0=theta0,
        thetaL=thetaL,
        thetaU=thetaU,
        noise_estim=False,
        nugget=1e-6,
        optimizer="BFGS",
        wait_iter=5,
        random_start=5 * dim,
        likelihood="concentrated",
        eval_budget=100 * dim,
    )

    return BO(
        search_space=space,
        obj_fun=obj_fun,
        model=model,
        DoE_size=dim * 5,
        max_FEs=max_FEs,
        verbose=False,
        n_point=1,
        minimize=True,
        ftarget=ftarget,
        logger=logfile,
    )


if __name__ == "__main__":
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    dims = (2, 5)
    fIDs = range(25)
    instance = range(1, size + 1)

    algorithms = [test_BO]

    opts = {
        "max_FEs": "100  * dim",
        "lb": -5,
        "ub": 5,
        "data_path": "",
    }
    opts["bbob_opt"] = {
        "comments": "max_FEs={0}".format(opts["max_FEs"]),
    }

    for algorithm in algorithms:
        opts["data_path"] = "./bbob_data/{}".format(algorithm.__name__)
        opts["bbob_opt"]["algid"] = algorithm.__name__
        for dim in dims:
            for fID in fIDs:
                run_optimizer(algorithm, dim, fID, instance[rank], logfile="./log", **opts)



import sys

import numpy as np

sys.path.insert(0, "./")

from bayes_optim import BO, Real, SearchSpace
from bayes_optim.surrogate import RandomForest

np.random.seed(123)


def fitness(x):
    x = np.asarray(x)
    return np.sum(x ** 2)


space = SearchSpace(
    [
        Real([0, 10], scale="log", precision=2),
        Real([0, 10], scale="log10", precision=2),
        Real([0, 1], scale="logit", precision=2),
        Real([-10, 10], scale="bilog", precision=2),
    ]
)
space.sample(1)

model = RandomForest(levels=space.levels)
opt = BO(
    search_space=space,
    obj_fun=fitness,
    model=model,
    DoE_size=5,
    max_FEs=50,
    verbose=True,
    n_point=1,
    acquisition_optimization={"optimizer": "OnePlusOne_Cholesky_CMA"},
)
print(opt.run())



import sys

import numpy as np
import pytest

sys.path.insert(0, "../")
from bayes_optim import BO, IntegerSpace, RealSpace
from bayes_optim._exception import ConstraintEvaluationError
from bayes_optim.search_space import DiscreteSpace
from bayes_optim.surrogate import GaussianProcess, RandomForest


def obj_fun2(x):
    return (x["pc"] - 0.2) ** 2 + x["mu"] + x["lambda"] + np.abs(x["p"] - 0.7)


def obj_fun(x):
    return np.sum(np.array(x) ** 2) + 5 * np.sum(np.array(x)) + 10


def h(x):
    return np.sum(x) - 1


def g(x):
    return [-x["pc"], x["mu"] - 1.9]


def test_BO_equality():
    dim = 2
    search_space = RealSpace([0, 1]) * dim
    thetaL = 1e-5 * np.ones(dim)
    thetaU = np.ones(dim)
    theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL
    model = GaussianProcess(
        corr="squared_exponential",
        theta0=theta0,
        thetaL=thetaL,
        thetaU=thetaU,
        nugget=1e-1,
        random_state=42,
    )
    xopt, _, __ = BO(
        search_space=search_space,
        obj_fun=obj_fun,
        eq_fun=h,
        model=model,
        max_FEs=20,
        DoE_size=3,
        acquisition_fun="MGFI",
        acquisition_par={"t": 2},
        acquisition_optimization={"optimizer": "BFGS"},
        verbose=True,
        random_seed=42,
    ).run()
    assert np.isclose(h(xopt), 0, atol=1e-1)


def test_BO_constraints():
    search_space = (
        IntegerSpace([1, 10], var_name="mu")
        + IntegerSpace([1, 10], var_name="lambda")
        + RealSpace([0, 1], var_name="pc")
        + RealSpace([0.005, 0.5], var_name="p")
    )
    model = RandomForest(levels=search_space.levels)
    xopt, _, __ = BO(
        search_space=search_space,
        obj_fun=obj_fun2,
        ineq_fun=g,
        model=model,
        max_FEs=10,
        DoE_size=3,
        eval_type="dict",
        acquisition_fun="MGFI",
        acquisition_par={"t": 2},
        n_job=1,
        n_point=1,
        verbose=True,
        random_seed=42,
    ).run()
    assert isinstance(xopt, dict)
    assert all(np.array(g(xopt)) <= 0)


def test_BO_bad_constraints():
    search_space = (
        DiscreteSpace(["1", "2", "3"], var_name="lambda")
        + RealSpace([0, 1], var_name="pc")
        + RealSpace([0.005, 0.5], var_name="p")
    )
    model = RandomForest(levels=search_space.levels)
    with pytest.raises(ConstraintEvaluationError):
        BO(
            search_space=search_space,
            obj_fun=lambda x: 10 * (x[0] == "3") + x[1] * x[2],
            ineq_fun=lambda x: sum(np.array(x) ** 2),
            model=model,
            max_FEs=10,
            DoE_size=3,
            eval_type="list",
            acquisition_fun="MGFI",
            acquisition_par={"t": 2},
            n_job=1,
            n_point=1,
            verbose=True,
            random_seed=42,
        ).run()



import os
import sys

sys.path.insert(0, "../")

import dill
import numpy as np
from bayes_optim import Solution


def test_1D():
    s = Solution(np.random.randint(0, 100, size=(5,)), index=1)
    assert isinstance(s[0], int)
    s.fitness = 3
    assert s.ndim == 1
    assert s.fitness == 3
    assert all(s.unique() == s)


def test_2D():
    # test for 2D solution
    A, B = np.random.randn(5, 3).tolist(), ["simida", "niubia", "bang", "GG", "blyat"]
    s = Solution([A[i] + [B[i]] for i in range(5)], verbose=True, fitness=0, fitness_name="f")
    assert s.ndim == 2
    assert s.N == 5
    assert s.dim == 4

    s[:, 0] = np.asarray(["wa"] * 5).reshape(-1, 1)
    assert np.all(s[:, 0] == "wa")

    s[3].fitness = 12
    assert s.fitness[3] == 12

    a = s[0]
    a.fitness = 3
    assert s.fitness[0] == 3

    s[2:4].fitness = 1
    assert np.all(s.fitness[2:4] == 1)

    s[0].index = "5"
    assert s.index[0] == "5"

    print(s[0:1])
    print(s[0, 0:3])
    print(s[:, 0])
    print(s[0:2][0, 0:2])
    print(s[0][0:2])

    print(s + s[3:5])
    print(s[0:2] + s[3:5])


def test_to_dict():
    s = Solution(np.random.randn(10, 5))
    print(s.to_dict(orient="index"))

    s = Solution(np.random.randn(10))
    s.to_dict()


def test_pickling():
    s = Solution(np.random.randn(10, 5))
    a = dill.dumps(s)
    s2 = dill.loads(a)
    assert np.all(s == s2)


def test_to_csv():
    # # test saving to csv
    s = Solution(np.random.randn(10, 5))
    s.to_csv("test.csv", header=True, attribute=True, index=True)
    os.remove("test.csv")


def test_from_dict():
    s = Solution(np.random.randn(10, 5))
    s_ = s.to_dict(orient="index")
    s2 = Solution.from_dict(s_)
    assert np.all(s == s2)
    assert isinstance(s.to_dict(orient="index", with_index=True), dict)
    assert isinstance(s.to_dict(orient="var", with_index=True), dict)



import os
import sys

import numpy as np

sys.path.insert(0, "./")

from bayes_optim import BO, RealSpace
from bayes_optim.surrogate import GaussianProcess, trend

np.random.seed(123)
dim = 5
lb, ub = -1, 5


def fitness(x):
    x = np.asarray(x)
    return np.sum(x ** 2)


space = RealSpace([lb, ub]) * dim
thetaL = 1e-10 * (ub - lb) * np.ones(dim)
thetaU = 10 * (ub - lb) * np.ones(dim)
theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

model = GaussianProcess(
    theta0=theta0,
    thetaL=thetaL,
    thetaU=thetaU,
    nugget=1e-3,
    noise_estim=True,
    optimizer="BFGS",
    wait_iter=3,
    random_start=dim,
    likelihood="concentrated",
    eval_budget=100 * dim,
)

opt = BO(
    search_space=space,
    obj_fun=fitness,
    model=model,
    DoE_size=5,
    max_FEs=20,
    verbose=True,
    n_point=1,
    acquisition_optimization={"optimizer": "OnePlusOne_Cholesky_CMA"},
)
print(opt.run())



from __future__ import annotations

import functools
from copy import copy, deepcopy
from typing import Callable, Dict, List, Union

import numpy as np
from joblib import Parallel, delayed
from scipy.stats import rankdata
from sklearn.decomposition import PCA
from sklearn.metrics import mean_absolute_percentage_error, r2_score

from . import acquisition_fun as AcquisitionFunction
from .bayes_opt import BO, ParallelBO
from .search_space import RealSpace, SearchSpace
from .surrogate import GaussianProcess


class LinearTransform(PCA):
    def __init__(self, minimize: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.minimize = minimize

    def fit(self, X: np.ndarray, y: np.ndarray) -> LinearTransform:
        """center the data matrix and scale the data points with respect to the objective values

        Parameters
        ----------
        data : Solution
            the data matrix to scale

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            the scaled data matrix, the center data matrix, and the objective value
        """
        self.center = X.mean(axis=0)
        X_centered = X - self.center
        y_ = -1 * y if not self.minimize else y
        r = rankdata(y_)
        N = len(y_)
        w = np.log(N) - np.log(r)
        w /= np.sum(w)
        X_scaled = X_centered * w.reshape(-1, 1)
        return super().fit(X_scaled)  # fit the PCA transformation on the scaled data matrix

    def transform(self, X: np.ndarray) -> np.ndarray:
        return super().transform(X - self.center)  # transform the centered data matrix

    def fit_transform(self, X: np.ndarray, y: np.ndarray) -> np.ndarray:
        self.fit(X, y)
        return self.transform(X)

    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        if not hasattr(self, "components_"):
            return X
        return super().inverse_transform(X) + self.center


def penalized_acquisition(x, acquisition_func, bounds, pca, return_dx):
    bounds_ = np.atleast_2d(bounds)
    # map back the candidate point to check if it falls inside the original domain
    x_ = pca.inverse_transform(x)
    idx_lower = np.nonzero(x_ < bounds_[:, 0])[0]
    idx_upper = np.nonzero(x_ > bounds_[:, 1])[0]
    penalty = -1 * (
        np.sum([bounds_[i, 0] - x_[i] for i in idx_lower])
        + np.sum([x_[i] - bounds_[i, 1] for i in idx_upper])
    )

    if penalty == 0:
        out = acquisition_func(x)
    else:
        if return_dx:
            # gradient of the penalty in the original space
            g_ = np.zeros((len(x_), 1))
            g_[idx_lower, :] = 1
            g_[idx_upper, :] = -1
            # get the gradient of the penalty in the reduced space
            g = pca.components_.dot(g_)
            out = (penalty, g)
        else:
            out = penalty
    return out


class PCABO(BO):
    """Dimensionality reduction using Principle Component Decomposition (PCA)

    References

    [RaponiWB+20]
        Raponi, Elena, Hao Wang, Mariusz Bujny, Simonetta Boria, and Carola Doerr.
        "High dimensional bayesian optimization assisted by principal component analysis."
        In International Conference on Parallel Problem Solving from Nature, pp. 169-183.
        Springer, Cham, 2020.

    """

    def __init__(self, n_components: Union[float, int] = None, **kwargs):
        super().__init__(**kwargs)
        if self.model is not None:
            self.logger.warn(
                "The surrogate model will be created automatically by PCA-BO. "
                "The input argument `model` will be ignored"
            )
        assert isinstance(self._search_space, RealSpace)
        self.__search_space = deepcopy(self._search_space)  # the original search space
        self._pca = LinearTransform(
            n_components=n_components, svd_solver="full", minimize=self.minimize
        )

    @staticmethod
    def _compute_bounds(pca: PCA, search_space: SearchSpace) -> List[float]:
        C = np.array([(l + u) / 2 for l, u in search_space.bounds])
        radius = np.sqrt(np.sum((np.array([l for l, _ in search_space.bounds]) - C) ** 2))
        C = C - pca.mean_ - pca.center
        C_ = C.dot(pca.components_.T)
        return [(_ - radius, _ + radius) for _ in C_]

    def _create_acquisition(self, fun=None, par=None, return_dx=False, **kwargs) -> Callable:
        acquisition_func = super()._create_acquisition(
            fun=fun, par={} if par is None else par, return_dx=return_dx, **kwargs
        )
        # TODO: make this more general for other acquisition functions
        # wrap the penalized acquisition function for handling the box constraints
        return functools.partial(
            penalized_acquisition,
            acquisition_func=acquisition_func,
            bounds=self.__search_space.bounds,  # hyperbox in the original space
            pca=self._pca,
            return_dx=return_dx,
        )

    def pre_eval_check(self, X: List) -> List:
        return X

    @property
    def xopt(self):
        if not hasattr(self, "data"):
            return None
        fopt = self._get_best(self.data.fitness)
        self._xopt = self.data[np.where(self.data.fitness == fopt)[0][0]]
        return self._xopt

    def ask(self, n_point: int = None) -> List[List[float]]:
        return self._pca.inverse_transform(super().ask(n_point))

    def tell(self, new_X, new_y):
        self.logger.info(f"observing {len(new_X)} points:")
        for i, x in enumerate(new_X):
            self.logger.info(f"#{i + 1} - fitness: {new_y[i]}, solution: {x}")

        index = np.arange(len(new_X))
        if hasattr(self, "data"):
            index += len(self.data)
        # convert `new_X` to a `Solution` object
        new_X = self._to_geno(new_X, index=index, n_eval=1, fitness=new_y)
        self.iter_count += 1
        self.eval_count += len(new_X)

        new_X = self.post_eval_check(new_X)  # remove NaN's
        self.data = self.data + new_X if hasattr(self, "data") else new_X
        # re-fit the PCA transformation
        X = self._pca.fit_transform(np.array(self.data), self.data.fitness)
        bounds = self._compute_bounds(self._pca, self.__search_space)
        # re-set the search space object for the reduced (feature) space
        self._search_space = RealSpace(bounds)
        # update the surrogate model
        self.update_model(X, self.data.fitness)
        self.logger.info(f"xopt/fopt:\n{self.xopt}\n")

    def update_model(self, X: np.ndarray, y: np.ndarray):
        # NOTE: the GPR model will be created since the effective search space (the reduced space
        # is dynamic)
        dim = self._search_space.dim
        bounds = np.array(self._search_space.bounds)
        _range = bounds[:, 1] - bounds[:, 0]
        thetaL, thetaU = 1e-8 * _range, 10 * _range
        self.model = GaussianProcess(
            theta0=np.random.rand(dim) * (thetaU - thetaL) + thetaL,
            thetaL=thetaL,
            thetaU=thetaU,
            nugget=1e-8,
            noise_estim=False,
            random_start=dim,
            eval_budget=100 * dim,
        )

        _std = np.std(y)
        y_ = y if np.isclose(_std, 0) else (y - np.mean(y)) / _std

        self.fmin, self.fmax = np.min(y_), np.max(y_)
        self.frange = self.fmax - self.fmin

        self.model.fit(X, y_)
        y_hat = self.model.predict(X)

        r2 = r2_score(y_, y_hat)
        MAPE = mean_absolute_percentage_error(y_, y_hat)
        self.logger.info(f"model r2: {r2}, MAPE: {MAPE}")


class MultiAcquisitionBO(ParallelBO):
    """Using multiple acquisition functions for parallelization"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        assert self.n_point > 1

        self._acquisition_fun = "MGFI"
        self._acquisition_fun_list = ["MGFI", "UCB"]
        self._sampler_list = [
            lambda x: np.exp(np.log(x["t"]) + 0.5 * np.random.randn()),
            lambda x: 1 / (1 + np.exp((x["alpha"] * 4 - 2) + 0.6 * np.random.randn())),
        ]
        self._par_name_list = ["t", "alpha"]
        self._acquisition_par_list = [{"t": 1}, {"alpha": 0.1}]
        self._N_acquisition = len(self._acquisition_fun_list)

        for i, _n in enumerate(self._par_name_list):
            _criterion = getattr(AcquisitionFunction, self._acquisition_fun_list[i])()
            if _n not in self._acquisition_par_list[i]:
                self._acquisition_par_list[i][_n] = getattr(_criterion, _n)

    def _batch_arg_max_acquisition(self, n_point: int, return_dx: bool, fixed: Dict = None):
        criteria = []
        for i in range(n_point):
            k = i % self._N_acquisition
            _acquisition_fun = self._acquisition_fun_list[k]
            _acquisition_par = self._acquisition_par_list[k]
            _par = self._sampler_list[k](_acquisition_par)
            _acquisition_par = copy(_acquisition_par)
            _acquisition_par.update({self._par_name_list[k]: _par})
            criteria.append(
                self._create_acquisition(
                    fun=_acquisition_fun, par=_acquisition_par, return_dx=return_dx, fixed=fixed
                )
            )

        if self.n_job > 1:
            __ = Parallel(n_jobs=self.n_job)(
                delayed(self._argmax_restart)(c, logger=self.logger) for c in criteria
            )
        else:
            __ = [list(self._argmax_restart(_, logger=self.logger)) for _ in criteria]

        return tuple(zip(*__))



from __future__ import annotations

from typing import Any, List

# from py_expression_eval import Parser


class Node:
    r"""Node of a tree used to represent the conditional search space"""

    def __init__(self, name: str, data: Any = None):
        self.name: str = name
        self.data: Any = data
        self.is_root: bool = True  # is this node a root?
        self.children: List[Node] = []
        self.branches: List = []  # branching conditions

    def add_child(self, node: Node, branch: str = None) -> Node:
        node.is_root = False
        self.children.append(node)
        # TODO: add parsing exceptions here
        # expr = Parser().parse(branch) if branch else None
        self.branches.append(branch)
        return self

    def add_child_from_dict(self, d: dict, cache: dict = None) -> Node:
        if cache is None:
            cache = dict()
        children = d.get(self.name, None)
        if children:  # an internal node
            for info in children:
                key, branch = info["name"], info["condition"]
                self.add_child(
                    cache[key] if key in cache else Node.add_child_from_dict(Node(key), d, cache),
                    branch,
                )
            cache[self.name] = self
        return self

    @classmethod
    def from_dict(cls, d: dict) -> List[Node]:
        cache = dict()
        for k in d.keys():
            if k in cache:
                continue
            cls(k).add_child_from_dict(d, cache)
        return [node for _, node in cache.items() if node.is_root]

    def pprint(self, _prefix="", branch=None, _last=True):
        print(
            _prefix,
            "`- " if _last else "|- ",
            f"<{branch}> - " if branch else "",
            self.name,
            sep="",
        )
        _prefix += "   " if _last else "|  "
        child_count = len(self.children)
        for i, child in enumerate(self.children):
            _last = i == (child_count - 1)
            Node.pprint(child, _prefix, self.branches[i], _last)

    def eval_path(self, path: List) -> List[Node]:
        pass

    # def get_all_path(self) -> List[List]:
    #     if not self.children:
    #         return [[]]

    #     path = []
    #     # idx = {}
    #     # `set` for getting unique branches
    #     # d = {name: i for i, name in enumerate(set(self.branches))}
    #     # for key in self.branches:
    #     #     idx.setdefault(key, []).append(d[key])

    #     for br in set(self.branches):
    #         idx = [i for i, b in enumerate(self.branches) if b == br]
    #         _ = [Node.get_all_path(self.children[i]) for i in idx]
    #         _path = sum(_, [[]])
    #         breakpoint()
    #         path += [[br].append(p) if p is not None else [br] for p in _path]

    #     return path

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name



import json
import re
import sys

sys.path.insert(0, "../")

import numpy as np
import pytest
from bayes_optim.search_space import (
    Discrete,
    DiscreteSpace,
    Integer,
    IntegerSpace,
    Node,
    Real,
    RealSpace,
    SearchSpace,
    Subset,
    SubsetSpace,
    Variable,
)
from bayes_optim.solution import Solution


def test_Variable():
    x = Real([1e-20, 5], "x", 2, scale="log10")
    bounds_transformed = getattr(x, "_bounds_transformed")
    assert bounds_transformed[0] == -20
    assert np.isclose(bounds_transformed[1], 0.6989700043360189)

    x = IntegerSpace([[5, 10]] * 3, ["x", "y", "z"])
    assert all(np.asarray(x.var_name) == np.asarray(["x", "y", "z"]))
    assert all(np.asarray(x.var_type) == np.asarray(["Integer"] * 3))


def test_real_warning():
    x = Real([-np.inf, 5], "x", 2, scale="log10")
    assert x.bounds[0] == 1e-300


def test_SearchSpace_var_name():
    cs = (
        RealSpace([1e-10, 1e-1], "x", 1e-3, scale="log")
        + IntegerSpace([-10, 10], "y")
        + DiscreteSpace(["A", "B", "C", "D", "E"], "z")
    )
    cs.var_name = "x"
    assert all(np.asarray(cs.var_name) == np.asarray(["x0", "x1", "x2"]))

    cs.var_name = ["y0", "y1", "y2"]
    assert all(np.asarray(cs.var_name) == np.asarray(["y0", "y1", "y2"]))

    with pytest.raises(AssertionError):
        cs.var_name = ["y0", "y1"]


def test_eq_ne():
    cs = RealSpace([[1e-10, 1e-1]] * 2, "x", 0.01, scale="log")
    cs2 = RealSpace([1e-10, 1e-1], "x", 0.01, scale="log") * 2
    assert cs.data[0] == cs2.data[0]
    assert cs.data[0] in cs2.data
    assert cs == cs2


def test_contains():
    cs = (
        IntegerSpace([-10, 10], "y")
        + DiscreteSpace(["A", "B", "C", "D", "E"], "z")
        + RealSpace([1e-10, 1e-1], "x", 0.01, scale="log")
    )
    assert RealSpace([1e-10, 1e-1], "x", 0.01, scale="log") in cs


def test_in():
    cs = (
        IntegerSpace([-10, 10], "y")
        + DiscreteSpace(["A", "B", "C", "D", "E"], "z")
        + RealSpace([1e-10, 1e-1], "x", 0.01, scale="log")
    )
    x = Solution(cs.sample(1)[0], var_name=cs.var_name)
    assert RealSpace([1e-10, 1e-1], "x", 0.01, scale="log") in cs
    assert "x" in cs
    assert "xx" not in cs
    assert x.tolist() in cs
    assert x.to_dict() in cs


def test_sample_with_constraints():
    g = lambda x: x - 0.1
    cs = RealSpace([1e-10, 1e-1], "x", 0.01, scale="log")
    X = cs.sample(10, g=g)
    assert all(list(map(lambda x: g(x) <= 0, X)))


def test_SearchSpace_remove():
    cs = (
        RealSpace([1e-10, 1e-1], "x", 1e-3, scale="log")
        + IntegerSpace([-10, 10], "y")
        + DiscreteSpace(["A", "B", "C", "D", "E"], "z")
    )

    cs.remove(0)
    assert isinstance(cs[[0]], IntegerSpace)

    cs.remove("z")
    assert isinstance(cs, IntegerSpace)

    with pytest.raises(KeyError):
        cs.remove("aaa")


def test_SearchSpace_mul():
    cs = (
        RealSpace([0, 5], "x") + IntegerSpace([-10, 10], "y") + DiscreteSpace(["A", "B", "C"], "z")
    )
    __ = ["x0", "y0", "z0", "x1", "y1", "z1"]
    assert (cs * 2).dim == 6
    assert all(np.array((2 * cs).var_name) == np.asarray(__))

    cs *= 2
    assert cs.dim == 6


def test_SearchSpace_sub():
    for _ in range(3):
        cs = (
            RealSpace([0, 5], "x")
            + IntegerSpace([-10, 10], "y")
            + DiscreteSpace(["A", "B", "C"], "z")
        )
        _cs = DiscreteSpace(["A", "B", "C"], "z") + IntegerSpace([10, 20], "p")

        cs2 = cs - cs[1:]
        assert cs2.var_name[0] == "x"
        assert cs.dim == 3

        assert set((cs - _cs).var_name) == set(["x", "y"])
        cs -= _cs
        assert set(cs.var_name) == set(["x", "y"])

        cs -= cs[1:]
        assert cs.var_name[0] == "x"


def test_SearchSpace_concat():
    cs_list = [RealSpace([0, 5], "x") for _ in range(3)]
    cs = SearchSpace.concat(*cs_list)
    assert all(np.array(cs.var_name) == np.array(["x0", "x1", "x2"]))
    assert isinstance(cs, RealSpace)

    cs_list = [RealSpace([0, 5], "x"), DiscreteSpace([1, 2, 3]), IntegerSpace([-5, 5])]
    cs = SearchSpace.concat(*cs_list)
    assert isinstance(cs, SearchSpace)


def test_SearchSpace_iadd():
    cs = RealSpace([0, 5], "x")
    cs += RealSpace([5, 10])
    assert isinstance(cs, RealSpace)
    cs += IntegerSpace([-5, 5])
    assert isinstance(cs, SearchSpace)


def test_SearchSpace_slice():
    cs = (
        RealSpace([1, 5], "x", 2, scale="log")
        + IntegerSpace([-10, 10], "y")
        + DiscreteSpace(["A", "B", "C"], "z")
    )
    assert isinstance(cs[[0]], RealSpace)
    assert isinstance(cs[[1]], IntegerSpace)
    assert isinstance(cs[[2]], DiscreteSpace)

    assert isinstance(cs[0], Real)
    assert isinstance(cs[1], Integer)
    assert isinstance(cs[2], Discrete)
    assert isinstance(cs["z"], Discrete)

    cs = (
        RealSpace([1, 5], "x", 2, scale="log") * 2
        + IntegerSpace([-10, 10], "y")
        + DiscreteSpace(["A", "B", "C"], "z")
    )
    assert isinstance(cs[:2], RealSpace)
    assert isinstance(cs[["x0", "x1"]], RealSpace)
    assert isinstance(cs[[False, False, True, False]], IntegerSpace)

    assert isinstance(cs.filter(["x0", "x1"]), RealSpace)


def test_sample():
    cs = (
        RealSpace([1e-10, 1e-1], "x", 1e-3, scale="log")
        + IntegerSpace([-10, 10], "y")
        + DiscreteSpace(["A", "B", "C", "D", "E"], "z")
    )
    X = cs.sample(10)
    assert np.asarray(X).shape == (10, 3)
    cs.sample(5, method="LHS")
    X = cs.sample(5, method="uniform")
    cs.to_linear_scale(X)


def test_constraints():
    cs = RealSpace([-5, 5], "x") * 2
    g = lambda x: x[0] + x[1] - 5
    X = cs.sample(10, g=g)
    assert all([g(x) <= 0 for x in X])

    X = cs.sample(10, g=lambda x: x[0] + 5.1)
    assert len(X) == 0


def test_scale():
    cs = RealSpace([1e-10, 1e-1], "x", scale="log", random_seed=42)
    x = cs.sample(1)
    assert np.isclose(x, 2.3488813e-07)
    assert np.isclose(cs.to_linear_scale(-15.812834391811666), 1.35697948e-07)
    assert np.isclose(cs.to_linear_scale([-15.812834391811666]), 1.35697948e-07)
    assert np.isclose(cs.to_linear_scale((-15.812834391811666)), 1.35697948e-07)

    C = RealSpace([1, 5], scale="log")
    assert getattr(C.data[0], "_bounds_transformed")[0] == 0

    C = RealSpace([0.5, 0.8], scale="logit")
    assert getattr(C.data[0], "_bounds_transformed")[0] == 0

    C = RealSpace([-1, 1], scale="bilog")
    assert getattr(C.data[0], "_bounds_transformed")[0] == -np.log(2)
    assert getattr(C.data[0], "_bounds_transformed")[1] == np.log(2)

    C = RealSpace([-1, 1], scale="bilog") * 2
    x = C.to_linear_scale([-np.log(2), np.log(2)])
    assert np.all(x == np.array([-1, 1]))


def test_precision():
    cs = RealSpace([0, 1], precision=2) * 3
    X = cs.sample(1, method="LHS")
    X = [re.sub(r"^-?\d+\.(\d+)$", r"\1", str(_)) for _ in X[0]]
    assert all([len(x) <= 2 for x in X])

    X = cs.round(np.random.randn(3))
    X = [re.sub(r"^-?\d+\.(\d+)$", r"\1", str(_)) for _ in X[0]]
    assert all([len(x) <= 2 for x in X])

    X = np.random.rand(2, 3)
    assert isinstance(cs.round(X), np.ndarray)

    X = Solution(cs.sample(10, method="LHS"))
    cs.round(X)

    cs = (
        RealSpace([0, 1], "x", precision=2)
        + IntegerSpace([-10, 10], "y")
        + DiscreteSpace(["A", "B", "C", "D", "E"], "z")
    )

    X = cs.sample(1, method="LHS")[0][0]
    X = re.sub(r"^-?\d+\.(\d+)$", r"\1", str(X))
    assert len(X) <= 2


def test_iter():
    cs = (
        RealSpace([1e-10, 1e-1], "x", 1e-3, scale="log")
        + IntegerSpace([-10, 10], "y")
        + DiscreteSpace(["A", "B", "C", "D", "E"], "z")
    )
    cs *= 2

    for var in iter(cs):
        assert isinstance(var, Variable)


def test_from_dict():
    cs = SearchSpace.from_dict(
        {
            "activation": {
                "type": "c",
                "range": [
                    "elu",
                    "selu",
                    "softplus",
                    "softsign",
                    "relu",
                    "tanh",
                    "sigmoid",
                    "hard_sigmoid",
                    "linear",
                ],
                "N": 3,
            }
        }
    )

    assert cs.dim == 3
    assert cs.var_name[0] == "activation0"

    with open("./shiny/example.json") as f:
        data = json.load(f)

    cs = SearchSpace.from_dict(data["search_param"])


def test_update():
    cs = RealSpace([0, 5], "x") * 3
    cs2 = RealSpace([-100, 100], "x1") + IntegerSpace([0, 10], "y")
    cs.update(cs2)
    assert "y" in cs.var_name
    assert cs[[1]].data[0].bounds[0] == -100
    assert cs[[1]].data[0].bounds[1] == 100


def test_filter():
    cs = (
        RealSpace([1e-10, 1e-1], "x", 0.01, scale="log")
        + IntegerSpace([-10, 10], "y")
        + DiscreteSpace(["A", "B", "C", "D", "E"], "z")
    )
    cs *= 2
    assert cs.filter(["x1"]).var_name == ["x1"]
    assert "x1" not in cs.filter(["x1"], invert=True).var_name


def test_subset():
    x = SubsetSpace(["a", "b", "c", "d"]) * 2
    x.sample(10)
    xx = Subset(["a", "b", "c", "d"])
    xx.sample(10)


def test_node():

    info = {
        "ccc2": [{"name": "cccc", "condition": "ccc2 == 2"}],
        "c1": [
            {"name": "cc1", "condition": "c1 == 1"},
            {"name": "cc2", "condition": "c1 == 10"},
        ],
        "p1": [
            {"name": "c1", "condition": "p1 == 1"},
            {"name": "c2", "condition": "p1 == 2"},
            {"name": "c3", "condition": "p1 == 3"},
        ],
        "cc2": [
            {"name": "ccc1", "condition": "cc2 == 1"},
            {"name": "ccc2", "condition": "cc2 == 2"},
        ],
        "p2": [{"name": "a", "condition": "p2 == 'a'"}, {"name": "b", "condition": "p2 == 'b'"}],
    }
    root = Node.from_dict(info)
    assert root[0].name == "p1"
    assert root[1].name == "p2"

    # print(root[0].get_all_path())


# def test_condition():
#     import string
#     v = Real([0, 5], "x", conditions="y == 'A'")
#     breakpoint()
#     cs = SearchSpace(
#         [Real([0, 5], "x", conditions="`y` == 'A'"), Discrete(list(string.ascii_uppercase), "y")]
#     )



import sys

import numpy as np

sys.path.insert(0, "../")
from bayes_optim import fmin

np.random.seed(42)


def test_fmin():
    def f(x):
        x = np.asarray(x)
        return np.sum(x ** 2)

    minimum = fmin(f, [-5] * 2, [5] * 2, seed=42, max_FEs=30, verbose=False)
    assert len(minimum) == 5
    assert len(minimum[0]) == 2
    # assert np.isclose(minimum[1], 0.007165794451494286)
    # assert all(np.isclose(minimum[0], [-0.04300030341296269, 0.07291617350003657]))

    # test warm starting
    X = np.random.rand(10, 2) * 10 - 5
    y = [f(x) for x in X]
    minimum = fmin(f, [-5] * 2, [5] * 2, x0=X, y0=y, max_FEs=20, verbose=False)
    assert minimum[2] == 20

    minimum = fmin(f, [-5] * 2, [5] * 2, x0=X, max_FEs=5, verbose=False)
    assert minimum[2] == 5


# def test_multi_acquisition():
# dim_r = 2  # dimension of the real values
# def obj_fun(x):
#     x_r = np.array([x['continuous_%d'%i] for i in range(dim_r)])
#     x_i = x['ordinal']
#     x_d = x['nominal']
#     _ = 0 if x_d == 'OK' else 1
#     return np.sum(x_r ** 2) + abs(x_i - 10) / 123. + _ * 2

# search_space = RealSpace([-5, 5], var_name='continuous') * dim_r + \
#     IntegerSpace([5, 15], var_name='ordinal') + \
#     DiscreteSpace(['OK', 'A', 'B', 'C', 'D', 'E', 'F', 'G'], var_name='nominal')

# model = RandomForest(levels=search_space.levels)

# opt = MultiAcquisitionBO(
#     search_space=search_space,
#     obj_fun=obj_fun,
#     model=model,
#     max_FEs=8,
#     DoE_size=4,    # the initial DoE size
#     eval_type='dict',
#     n_job=4,       # number of processes
#     n_point=4,     # number of the candidate solution proposed in each iteration
#     verbose=True   # turn this off, if you prefer no output
# )

# xopt, fopt, stop_dict = opt.run()
# print('xopt: {}'.format(xopt))
# print('fopt: {}'.format(fopt))
# print('stop criteria: {}'.format(stop_dict))



import sys

import numpy as np
import pytest

sys.path.insert(0, "../")
from bayes_optim import MOBO
from bayes_optim._exception import RecommendationUnavailableError
from bayes_optim.search_space import (BoolSpace, DiscreteSpace, IntegerSpace,
                                      RealSpace)
from bayes_optim.surrogate import GaussianProcess, RandomForest

np.random.seed(123)


def f1(x):
    return (
        (x["continuous"] - 3) ** 2
        + abs(x["ordinal"] + 10) / 123.0
        + (x["nominal"] != "OK") * 2
        + int(x["bool"]) * 3
    )


def f2(x):
    return (
        x["continuous"] ** 3
        + abs(x["ordinal"] - 10) / 123.0
        + (x["nominal"] != "G") * 2
        + int(not x["bool"]) * 3
    )


def test_3D():
    search_space = (
        RealSpace([0, 100], var_name="Kp", precision=2)
        + RealSpace([0, 100], var_name="Ki", precision=2)
        + RealSpace([0, 100], var_name="Kd", precision=2)
    )
    f1 = lambda x: x["Kp"] ** 2 + x["Ki"] + x["Kd"] ** 2
    f2 = lambda x: x["Kp"] + x["Ki"] ** 2 + x["Kd"] ** 2
    f3 = lambda x: x["Kp"] ** 2 + x["Ki"] + x["Kd"]
    dim = search_space.dim
    thetaL = 1e-10 * 100 * np.ones(dim)
    thetaU = 10 * 100 * np.ones(dim)
    theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

    model = GaussianProcess(
        theta0=theta0,
        thetaL=thetaL,
        thetaU=thetaU,
        nugget=0,
        noise_estim=False,
        likelihood="concentrated",
    )
    opt = MOBO(
        search_space=search_space,
        obj_fun=(f1, f2, f3),
        model=model,
        max_FEs=100,
        DoE_size=1,  # the initial DoE size
        eval_type="dict",
        n_job=1,  # number of processes
        verbose=True,  # turn this off, if you prefer no output
        minimize=True,
        acquisition_optimization={"optimizer": "OnePlusOne_Cholesky_CMA"},
    )
    for _ in range(100):
        X = opt.ask(1)
        opt.tell(X, [(f1(x), f2(x), f3(x)) for x in X])

    opt.recommend()
    with pytest.raises(NotImplementedError):
        X = opt.ask(3)


def test_with_constraints():
    search_space = (
        RealSpace([0, 100], var_name="left", precision=2)
        + RealSpace([0, 100], var_name="up", precision=2)
        + RealSpace([0, 100], var_name="right", precision=2)
    )
    f1 = lambda x: x["left"] ** 2 + x["up"] + x["right"] ** 2
    f2 = lambda x: 10 * x["left"] - x["up"] ** 2 + x["right"]
    g = lambda x: x["left"] + x["up"] + x["right"] - 100
    dim = search_space.dim
    thetaL = 1e-10 * 100 * np.ones(dim)
    thetaU = 10 * 100 * np.ones(dim)
    theta0 = np.random.rand(dim) * (thetaU - thetaL) + thetaL

    model = GaussianProcess(
        theta0=theta0,
        thetaL=thetaL,
        thetaU=thetaU,
        nugget=0,
        noise_estim=False,
        likelihood="concentrated",
    )
    opt = MOBO(
        search_space=search_space,
        obj_fun=(f1, f2),
        ineq_fun=g,
        model=model,
        max_FEs=100,
        DoE_size=1,  # the initial DoE size
        eval_type="dict",
        n_job=1,  # number of processes
        verbose=True,  # turn this off, if you prefer no output
        acquisition_optimization={"optimizer": "OnePlusOne_Cholesky_CMA"},
    )
    for _ in range(50):
        X = opt.ask(1)
        opt.tell(X, [(f1(x), f2(x)) for x in X])
        assert g(X[0]) <= 0

    opt.recommend()
    with pytest.raises(NotImplementedError):
        X = opt.ask(3)


def test_fixed_var():
    search_space = (
        BoolSpace(var_name="bool")
        + IntegerSpace([5, 15], var_name="ordinal")
        + RealSpace([-5, 5], var_name="continuous", precision=2)
        + DiscreteSpace(["OK", "A", "B", "C", "D", "E", "F", "G"], var_name="nominal")
    )
    opt = MOBO(
        search_space=search_space,
        obj_fun=(f1, f2),
        model=RandomForest(levels=search_space.levels),
        max_FEs=100,
        DoE_size=1,  # the initial DoE size
        eval_type="dict",
        n_job=1,  # number of processes
        verbose=True,  # turn this off, if you prefer no output
    )
    X = opt.ask(3, fixed={"ordinal": 5, "continuous": 3.2})
    assert all([x["ordinal"] == 5 and x["continuous"] == 3.2 for x in X])
    opt.tell(X, [(f1(x), f2(x)) for x in X])
    X = opt.ask(1, fixed={"nominal": "OK", "bool": False})
    assert all([x["nominal"] == "OK" and not x["bool"] for x in X])
    opt.recommend()

    with pytest.raises(NotImplementedError):
        X = opt.ask(3)


def test_recommend():
    search_space = (
        RealSpace([10, 30], var_name="p1", precision=2)
        + IntegerSpace([20, 40], var_name="p2")
        + DiscreteSpace([128, 256, 512], var_name="p3")
        + BoolSpace(var_name="p4")
    )
    opt = MOBO(
        search_space=search_space,
        obj_fun=(f1, f2),
        model=RandomForest(levels=search_space.levels),
        max_FEs=100,
        DoE_size=3,  # the initial DoE size
        eval_type="dict",
        n_job=1,  # number of processes
        verbose=True,  # turn this off, if you prefer no output
    )
    with pytest.raises(RecommendationUnavailableError):
        opt.recommend()


def test_constraint():
    search_space = (
        BoolSpace(var_name="bool")
        + IntegerSpace([5, 15], var_name="ordinal")
        + RealSpace([-5, 5], var_name="continuous")
        + DiscreteSpace(["OK", "A", "B", "C", "D", "E", "F", "G"], var_name="nominal")
    )
    opt = MOBO(
        search_space=search_space,
        obj_fun=(f1, f2),
        model=RandomForest(levels=search_space.levels),
        ineq_fun=lambda x: x["continuous"],
        max_FEs=10,
        DoE_size=3,  # the initial DoE size
        eval_type="dict",
        n_job=1,  # number of processes
        verbose=True,  # turn this off, if you prefer no output
    )
    opt.run()
